#ifndef _SNAPSHOT_H
#define _SNAPSHOT_H

#include <linux/ip.h>

/*
 * should be synchronized with kernel ipt_SNAPSHOT.h !!!
 */

#define IPPROTO_SNAPSHOT             240
#define IPOPT_SNAPSHOT               100

// snapshot message type (1..255)
#define SNAPSHOT_MSGTYPE_RESET        1
#define SNAPSHOT_MSGTYPE_CLEAR        2
#define SNAPSHOT_MSGTYPE_CONTROL      3
#define SNAPSHOT_MSGTYPE_VMAP         4
#define SNAPSHOT_MSGTYPE_RMAP         5
#define SNAPSHOT_MSGTYPE_MARKER       6
#define SNAPSHOT_MSGTYPE_SNAPSHOT     7

#define SNAPSHOT_MSGTYPE_CONTROL_ACK  (SNAPSHOT_MSGTYPE_CONTROL | 0x80)
#define SNAPSHOT_MSGTYPE_VMAP_ACK     (SNAPSHOT_MSGTYPE_VMAP    | 0x80)
#define SNAPSHOT_MSGTYPE_RMAP_ACK     (SNAPSHOT_MSGTYPE_RMAP    | 0x80)

#define SNAPSHOT_MSGTYPE_SMAP         10
#define SNAPSHOT_MSGTYPE_STAT         11
#define SNAPSHOT_MSGTYPE_SMAP_ACK     (SNAPSHOT_MSGTYPE_SMAP    | 0x80)

#define SNAPSHOT_OPTLEN (sizeof(struct snapshot_opt))
#define SNAPSHOT_IPHLEN (sizeof(struct iphdr) + SNAPSHOT_OPTLEN)

#define MAX_LINK_LIMIT                50
#define MAX_STAT_LIMIT                100

// snapshot IP option
struct snapshot_opt {
  __u8 type;            // IPOPT_SNAPSHOT
  __u8 length;          // SNAPSHOT_OPTLEN
  __u8 msgtype;         // SNAPSHOT_MSGTYPE_XXX
  __u8 padding;         // zero-padding
  __u32 seqno;
};

// snapshot message header of VMAP
struct snapshot_msghdr_vmap {
  __u16 entrycnt;
  __u16 snapshotcnt;
};

// snapshot message header of RMAP
struct snapshot_msghdr_rmap {
  __u16 entrycnt;
  __u8 connected;       // 1: direct-connected to victim, 0: otherwise
  __u8 padding;
};

// snapshot message header of SMAP
struct snapshot_msghdr_smap {
  __u16 entrycnt;
  __u16 padding;
};

// snapshot message header of SNAPSHOT
struct snapshot_msghdr_snapshot {
  __u16 entrycnt;
  __u16 padding;
};

// snapshot message header of STAT
struct snapshot_msghdr_stat {
  __u16 entrycnt;
  __u16 padding;
};

#endif /* _SNAPSHOT_H */
