#ifndef _SSNODE_H
#define _SSNODE_H

typedef struct {
  /*__u32*/ long long value[2];
} dual_t;

// array of ssnode
typedef struct ssnode_map_t {
  struct ssnode_t *node;
  int snapshotcnt;
  int cnt;
  int dual_idx;                   // current index of dual_t type data (0 or 1)

  dual_t acc_channel_state;       // deta^(k)
  /*__u32*/ long long victim_traffic;           // I_v(t1, t2)

  int unmatched_cnt;
} ssnode_map_t;

// array of ssnode ptr
typedef struct ssnode_link_t {
  struct ssnode_t **nptr;
  int cnt;

  dual_t *channel_state;          // channel state associated with its parent H_ip(t, t')
} ssnode_link_t;

typedef struct ssnode_t {
  __u32 addr;

  /*__u32*/ long long local_stat_traffic;       // local traffic by statistic
  /*__u32*/ long long local_traffic;            // local traffic L_i(t1, t2)
  /*__u32*/ long long local_traffic_upper;      // upper bound of L_i
  /*__u32*/ long long local_traffic_lower;      // lower bound of L_i

  dual_t local_state;             // local state C_i(t)
  dual_t down_channel_state_sum;  // sum of channel state from downstream routers D(R_i)
  dual_t acc_local_traffic;       // accumulative local traffic N_i(t)
  struct ssnode_link_t iup;       // map of immediate upstream routers
} ssnode_t;

inline int ssnode_alloc_map(ssnode_map_t *map, int cnt);
inline void ssnode_free_map(ssnode_map_t *map);
inline int ssnode_alloc_link(ssnode_link_t *link, int cnt);
inline void ssnode_free_link(ssnode_link_t *link);
inline int ssnode_node_index(ssnode_map_t *map, __u32 addr);
int ssnode_insert_links(ssnode_map_t *map, __u32 parent_addr,
                        __u32* link_addr_map, int link_addr_cnt);
void ssnode_free_map_all(ssnode_map_t *map);
int ssnode_insert_nodes(ssnode_map_t *map, __u32* node_addr_map,
                        int node_addr_cnt);
void ssnode_print_dual_map(ssnode_map_t *map);
void ssnode_print_map(ssnode_map_t *map, int dual_idx);
void ssnode_print_map_to_files(ssnode_map_t *map, int seq, FILE **fp);
void ssnode_print_cutoff_to_file(ssnode_map_t *map, double cutoff_value, const char *filename);
void ssnode_analyse(ssnode_map_t *map, int dual_idx);
int ssnode_parse_snapshot(ssnode_map_t *map, unsigned char *data, int datalen);
int ssnode_parse_stat(ssnode_map_t *map, unsigned char *data, int datalen);

#endif /* _SSNODE_H */
