#ifndef _DECQUEUE_H
#define _DECQUEUE_H

#include <asm/types.h>

typedef struct {
  unsigned char *buf;
  unsigned int size;
} dq_buffer_t;

typedef struct __dq_block dq_block_t;
typedef struct __dq_queue dq_queue_t;

struct __dq_block {
  unsigned int id;
  int nbuffer;
  unsigned int k_size, n_size, b_size;
  dq_buffer_t **buflist;
  unsigned int *index;
  dq_block_t *prev, *next;
  pthread_mutex_t mutex;
  dq_queue_t *que;
  struct timeval timestamp;
  pthread_mutex_t nbuffer_mutex;
  pthread_cond_t nbuffer_cond;
};

typedef struct {
  unsigned char **buf;
  unsigned int *index;
} dq_decinfo_t;

struct __dq_queue {
  int nblock;
  dq_block_t head;
  pthread_mutex_t mutex;
  unsigned int headid;
  unsigned int wsize;      // window size
  unsigned int k_size, n_size;
  __u32 src_addr, dst_addr;
  void *fec;
  pthread_mutex_t nblock_mutex;
  pthread_cond_t nblock_cond;
};

#define DQ_BLOCK_EXPIRYTIME (3000000)  // 1000 ms
//#define DQ_BLOCK_MAXWSIZE   1024

#define dq_lock_block(blk) pthread_mutex_lock(&(blk->mutex))
#define dq_unlock_block(blk) pthread_mutex_unlock(&(blk->mutex))
#define dq_lock_block_nbuffer(blk) pthread_mutex_lock(&(blk->nbuffer_mutex))
#define dq_unlock_block_nbuffer(blk) pthread_mutex_unlock(&(blk->nbuffer_mutex))

#define dq_lock_queue(que) pthread_mutex_lock(&(que->mutex))
#define dq_unlock_queue(que) pthread_mutex_unlock(&(que->mutex))
#define dq_lock_queue_nblock(que) pthread_mutex_lock(&(que->nblock_mutex))
#define dq_unlock_queue_nblock(que) pthread_mutex_unlock(&(que->nblock_mutex))

#define dq_isempty_queue(que) (que->nblock == 0)


#define dq_new_buffer(bufptr, size) dq_new_copy_buffer(bufptr, NULL, size)
int dq_new_copy_buffer(dq_buffer_t **bufptr, unsigned char *data,
                       unsigned int size);
int dq_new_buffer_link(dq_buffer_t **bufptr, unsigned char *link,
                       unsigned int size);
void dq_free_buffer(dq_buffer_t **bufptr);


int dq_new_block(dq_block_t **blkptr, unsigned int id, unsigned int k_size, 
                 unsigned int n_size, unsigned int b_size);
void dq_free_block(dq_block_t **blkptr);
void dq_show_block(dq_block_t *blk);
int dq_decode_block(dq_block_t *blk, dq_decinfo_t *dec, void *fec);
inline int dq_istimetoleave_block(dq_block_t *blk);


int dq_new_decinfo(dq_decinfo_t **decptr, unsigned int size);
void dq_free_decinfo(dq_decinfo_t **decptr);


int dq_new_queue(dq_queue_t **queptr, unsigned int headid, unsigned int wsize,
                 unsigned int k_size, unsigned int n_size,
                 __u32 src_addr, __u32 dst_addr);
void dq_free_queue(dq_queue_t **queptr);
void dq_show_queue(dq_queue_t *que);
void dq_show_queue_status(dq_queue_t *que);
inline int dq_isaccept_blockid(dq_queue_t *que, unsigned int id);



int dq_add_buffer(dq_block_t *blk, int idx, dq_buffer_t *buf);
dq_buffer_t *dq_del_buffer(dq_block_t *blk, int idx);


dq_block_t *dq_get_block(dq_queue_t *que, unsigned int id);
int dq_enqueue_block(dq_queue_t *que, dq_block_t *blk);
int dq_enqueue_create_block(dq_queue_t *que, dq_block_t **blkptr,
                            unsigned int id, unsigned int b_size);
dq_block_t *dq_dequeue_block(dq_queue_t *que);

#endif /* _DECQUEUE_H */
