#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <linux/ip.h>
#include <lib/util.h>
#include <lib/net/rawsock.h>

int rawsock_create_socket(rawsock_t *raw, __u8 ipproto)
{
  int one = 1;
  int sd = socket(PF_INET, SOCK_RAW, ipproto);
  if (sd < 0) {
    ERR_RET(sd);
    return sd;
  }

  if (setsockopt(sd, IPPROTO_IP, IP_HDRINCL, (char *) &one, one) == -1) {
    ERR_RET(-1);
  }

  raw->ipproto = ipproto;
  raw->sd = sd;
  return sd;
}

void rawsock_close_socket(rawsock_t *raw)
{
  close(raw->sd);
}

void rawsock_fill_info(rawsock_t *raw)
{
  int iphlen = (sizeof(struct iphdr) + raw->opt.optlen);

  raw->iph.version = 4;
  raw->iph.ihl = iphlen >> 2;
  raw->iph.tos = 0;
  raw->iph.tot_len = htons(iphlen + raw->data_len);
  raw->iph.id = htons(0);
  raw->iph.frag_off = htons(0);
  raw->iph.ttl = MAXTTL;
  raw->iph.protocol = raw->ipproto;
  raw->iph.check = 0;
  raw->iph.saddr = raw->src_addr.s_addr;
  raw->iph.daddr = raw->dst_addr.s_addr;
 
  raw->sockaddr.sin_family = AF_INET;
  raw->sockaddr.sin_addr.s_addr = raw->dst_addr.s_addr;
  raw->sockaddr.sin_port = htons(0);
}

void *rawsock_alloc_data(rawsock_t *raw, int data_len)
{
  raw->data_len = data_len;
  raw->data = malloc(raw->data_len);
  return raw->data;
}

void rawsock_free_data(rawsock_t *raw)
{
  if (raw->data != NULL) {
    free(raw->data);
    raw->data = NULL;
  }
}

int rawsock_send(rawsock_t *raw)
{
  struct msghdr msg;
  struct iovec iov[3];

  int retval;

  // IP header
  iov[0].iov_base = &(raw->iph);
  iov[0].iov_len = sizeof(struct iphdr);
  // IP options
  iov[1].iov_base = &(raw->opt);
  iov[1].iov_len = raw->opt.optlen;
  // IP payload
  iov[2].iov_base = raw->data;
  iov[2].iov_len = raw->data_len;
  msg.msg_iovlen = 3;

  msg.msg_name = &(raw->sockaddr);
  msg.msg_namelen = sizeof(struct sockaddr);
  msg.msg_iov = (void *) &iov;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;

  retval = sendmsg(raw->sd, &msg, 0);

  if (retval < 0) {
    ERR();
  }
  return retval;
}

int rawsock_recv(rawsock_t *raw)
{
  struct msghdr msg;
  struct iovec iov[3];

  int retval;

  // IP header
  iov[0].iov_base = &(raw->iph);
  iov[0].iov_len = sizeof(struct iphdr);
  // IP options
  iov[1].iov_base = &(raw->opt);
  iov[1].iov_len = raw->opt.optlen;
  // IP payload
  iov[2].iov_base = raw->data;
  iov[2].iov_len = raw->data_len;
  msg.msg_iovlen = 3;

  msg.msg_name = &(raw->sockaddr);
  msg.msg_namelen = sizeof(struct sockaddr);
  msg.msg_iov = (void *) &iov;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;

  retval = recvmsg(raw->sd, &msg, 0);

  if (retval < 0) {
    ERR();
  }
  return retval;
}
