#include <lib/wrapper/wrapsock.h>

int Socket(int domain, int type, int protocol, EXCEPTION *ex)
{
  int ret = socket(domain, type, protocol);
  return (ret == -1) ?
    (*ex)(-1, "socket: %s\n", strerror(errno)) : ret;
}

int Bind(int sockfd, struct sockaddr *my_addr, socklen_t addrlen,
         EXCEPTION *ex)
{
  int ret = bind(sockfd, my_addr, addrlen);
  return (ret == -1) ?
    (*ex)(-1, "bind: %s\n", strerror(errno)) : ret;
}

int Listen(int s, int backlog, EXCEPTION *ex)
{
  int ret = listen(s, backlog);
  return (ret == -1) ?
    (*ex)(-1, "listen: %s\n", strerror(errno)) : ret;  
}

int Accept(int s, struct sockaddr *addr, socklen_t *addrlen, EXCEPTION *ex)
{
  int ret = accept(s, addr, addrlen);
  return (ret == -1) ?
    (*ex)(-1, "accept: %s\n", strerror(errno)) : ret;  
}

int Connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen,
            EXCEPTION *ex)
{
  int ret = connect(sockfd, serv_addr, addrlen);
  return (ret == -1) ?
    (*ex)(-1, "connect: %s\n", strerror(errno)) : ret;    
}

int Getsockopt(int s, int level, int optname, void *optval, socklen_t *optlen,
               EXCEPTION *ex)
{
  int ret = getsockopt(s, level, optname, optval, optlen);
  return (ret == -1) ?
    (*ex)(-1, "getsockopt: %s\n", strerror(errno)) : ret;    
}

int Setsockopt(int s, int level, int optname, const void *optval,
               socklen_t optlen, EXCEPTION *ex)
{
  int ret = setsockopt(s, level, optname, optval, optlen);
  return (ret == -1) ?
    (*ex)(-1, "setsockopt: %s\n", strerror(errno)) : ret;    
}

int Send(int s, const void *msg, size_t len, int flags, EXCEPTION *ex)
{
  int ret = send(s, msg, len, flags);
  return (ret == -1) ?
    (*ex)(-1, "send: %s\n", strerror(errno)) : ret;
}

int Recv(int s, void *buf, size_t len, int flags, EXCEPTION *ex)
{
  int ret = recv(s, buf, len, flags);
  return (ret == -1) ?
    (*ex)(-1, "recv: %s\n", strerror(errno)) : ret;
}

