#include <linux/skbuff.h>
#include <net/checksum.h>
#include <linux/pkt_sched.h>
#include <linux/vmalloc.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include "ipsend.h"

// stolen from ip_output.c
static inline int output_maybe_reroute(struct sk_buff *skb)
{
  return skb->dst->output(skb);
}

static inline int ip_finish_output2(struct sk_buff *skb)
{
  struct dst_entry *dst = skb->dst;
  struct hh_cache *hh = dst->hh;
 
  if (hh) {
    read_lock_bh(&hh->hh_lock);
    memcpy(skb->data - 16, hh->hh_data, 16);
    read_unlock_bh(&hh->hh_lock);
    skb_push(skb, hh->hh_len);
    return hh->hh_output(skb);
  } else if (dst->neighbour)
    return dst->neighbour->output(skb);
 
  if (net_ratelimit())
    printk(KERN_DEBUG "ip_finish_output2: No header cache and no neighbour!\n");
  kfree_skb(skb);
  return -EINVAL;
}

static inline struct sk_buff*
ipsend_generate_skb(__u32 saddr, __u32 daddr, __u8 type, int datalen,
                    __u32 seqno)
{
  struct sk_buff *skb = NULL;
  struct iphdr *iph;
  struct rtable *rt;
  struct ipsend_opt *opt;
  int hh_len, ret;
  int totlen;

  // DEBUGP(">>> 1\n");

  // find route
  if ((ret = ip_route_output(&rt, daddr, saddr, 0, 0)))
  {
    DEBUGP("route failed: %d\n", ret);
    goto out;
	}
  
  if (rt->u.dst.dev == NULL) {
    DEBUGP("route failed: %d\n", ret);
    goto out;
  }

  // DEBUGP(">>> 2\n");
  
  // length
  totlen = IPSEND_IPHLEN + datalen;
  hh_len = (rt->u.dst.dev->hard_header_len + 15)&~15;
  
  // hh_len for headspace
  skb = alloc_skb(totlen + hh_len, GFP_ATOMIC);
  if (skb == NULL) {
    goto out;
  }

  // DEBUGP(">>> 3\n");
  
  skb_reserve(skb, hh_len);

  // DEBUGP(">>> 4\n");
  
  // assign device
  skb->dev = rt->u.dst.dev;
  skb->dst = dst_clone(&rt->u.dst);  // <- ?????

  // DEBUGP(">>> 5\n");
  
  skb->protocol = __constant_htons(ETH_P_IP);
  //skb->priority = TC_PRIO_MAX;
  skb->priority = TC_PRIO_BESTEFFORT;

  // DEBUGP(">>> 6\n");
  
  // build ethernet header
/*
  if (skb->dev->hard_header &&
  skb->dev->hard_header(skb, skb->dev, ETH_P_IP, skb->dev->dev_addr,
  skb->dev->dev_addr, skb->len) < 0)
  goto out_free;

  // DEBUGP(">>> 7\n");
  
	if (skb->dev->rebuild_header && skb->dev->rebuild_header(skb) < 0)
  goto out_free;

  // DEBUGP(">>> 8\n");
*/  
  skb->nh.iph = (struct iphdr *) skb_put(skb, totlen);

  // DEBUGP(">>> 9\n");
  
  // mark skb as altered
  skb->nfcache |= NFC_ALTERED;

  // default IP header
  iph = skb->nh.iph;
  iph->version = IPVERSION;
  iph->ihl = IPSEND_IPHLEN >> 2;
  iph->tos = 0;
  iph->tot_len = htons(totlen);
  iph->id = 0;
  iph->frag_off = 0;
  iph->ttl = MAXTTL;
  iph->saddr = saddr;
  iph->daddr = daddr;
  iph->protocol = IPPROTO_IPSEND;

  // DEBUGP(">>> 10\n");
  
  // IP options
  opt = IPSEND_OPTPTR(iph);
  opt->type = IPOPT_IPSEND;
  opt->length = IPSEND_OPTLEN;
  opt->msgtype = type;
  opt->padding = 0;
  opt->seqno = htonl(seqno);

  // DEBUGP(">>> 11\n");
  
  goto out;
  
out_free:
  // DEBUGP(">>> 12\n");
  kfree_skb(skb);
  skb = NULL;
  
out:
  return skb;
}

int ipsend_send_data(__u32 saddr, __u32 daddr, __u32 seqno, __u32 lastudp)
{
  struct sk_buff *skb;
  struct iphdr *iph;
  struct ipsend_opt *opt;
  __u8 *data;
  int datalen;
  int i;

  /*
  DEBUGP("send IPSEND [%x] from %u.%u.%u.%u to %u.%u.%u.%u\n",
         seqno, NIPQUAD(saddr), NIPQUAD(daddr));
  */
  datalen = 100;

  // generate default ipsend skb
  skb = ipsend_generate_skb(saddr, daddr, IPSEND_MSGTYPE_DATA,
                            datalen, 0);
  if (skb == NULL)
    return -1;

  iph = skb->nh.iph;

  opt = IPSEND_OPTPTR(iph);
  opt->seqno = htonl(seqno);
  
  // IP payload
  data = IPSEND_DATAPTR(iph);
  for (i=0; i<datalen; i++)
    data[i] = 0;

  ((__u32 *) data)[0] = htonl(lastudp);
  
  // DEBUGP(">>> 13\n");
    
  // calculate IP header checksum
  ip_send_check(iph);

  // DEBUGP(">>> 14\n");
  
  return NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, skb->dev,
                 ip_finish_output2);
/*
  // transmit buffer via device
  if (dev_queue_xmit(skb) < 0)
    return -1;

  // DEBUGP(">>> 15\n");
*/  
  return 0;
}
