#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <lib/util.h>
#include <lib/net/rawsock.h>
#include <router/ddos/edgemark/edgemark.h>

inline __u8 emmsg_msgtype(struct iphdr * iph, struct edgemark_opt * opt)
{
  if ((iph->ihl == (EDGEMARK_IPHLEN >> 2)) &&
      (opt->type == IPOPT_EDGEMARK) && (opt->length == EDGEMARK_OPTLEN)) {
    return opt->msgtype;
  }

  return 0;
}

void emmsg_dump_iphdr(struct iphdr *iph)
{
  struct in_addr in;

  printf("[ iphdr ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s 0x%04x\n"
         "%-15.15s 0x%02x\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s 0x%04x\n",
         "version", iph->version,
         "ihl", iph->ihl,
         "tos", iph->tos,
         "tot_len", ntohs(iph->tot_len),
         "id", ntohs(iph->id),
         "flags", ntohs(iph->frag_off) >> 13,
         "frag_off", ntohs(iph->frag_off) & 0x1fff,
         "ttl", iph->ttl,
         "protocol", iph->protocol, "checksum", ntohs(iph->check));

  in.s_addr = iph->saddr;
  printf("%-15.15s %s\n", "saddr", inet_ntoa(in));
  in.s_addr = iph->daddr;
  printf("%-15.15s %s\n", "daddr", inet_ntoa(in));
}

static const char *edgemark_msgtype_str(__u8 msgtype)
{
  switch (msgtype) {
  case EDGEMARK_MSGTYPE_MARK:
    return "MARK";
  case EDGEMARK_MSGTYPE_START:
    return "START";
  case EDGEMARK_MSGTYPE_STOP:
    return "STOP";
  }

  return "invalid";
}

void emmsg_dump_opt(struct edgemark_opt *opt)
{
  struct in_addr in;

  printf("[ edgemark_opt ]\n"
         "%-15.15s %-3d (%s)\n"
         "%-15.15s %-3d (%s)\n"
         "%-15.15s %-3d (%s)\n"
         "%-15.15s %d\n",
         "type", opt->type,
         (opt->type == IPOPT_EDGEMARK) ? "correct" : "wrong",
         "length", opt->length,
         (opt->length == EDGEMARK_OPTLEN) ? "correct" : "wrong",
         "msgtype", opt->msgtype,
         edgemark_msgtype_str(opt->msgtype),
         "mark_distance", opt->mark_distance);

  in.s_addr = opt->mark_start;
  printf("%-15.15s %s\n", "mark_start", inet_ntoa(in));

  in.s_addr = opt->mark_end;
  printf("%-15.15s %s\n", "mark_end", inet_ntoa(in));
}

int emmsg_send_control(__u32 localaddr, __u32 remoteaddr, __u8 msgtype)
{
  rawsock_t raw;
  struct edgemark_opt *opt;

  memset(&raw, 0, sizeof(rawsock_t));

  if (rawsock_create_socket(&raw, IPPROTO_RAW) < 0)
    ERR_RET(-1);

  // IP options
  opt = (struct edgemark_opt *) &(raw.opt);
  opt->type = IPOPT_EDGEMARK;
  opt->length = EDGEMARK_OPTLEN;
  opt->msgtype = msgtype;

  // fill up source and destination IP addresses
  raw.src_addr.s_addr = localaddr;
  raw.dst_addr.s_addr = remoteaddr;

  // complete packet info
  rawsock_fill_info(&raw);

  rawsock_send(&raw);

  close(raw.sd);

  return 0;
}
