#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <lib/util.h>
#include <lib/net/rawsock.h>
#include <router/ddos/snapshot/snapshot.h>
#include <router/ddos/snapshot/ssmsg.h>

// program information
const char *program_version = "0.1";
const char *program_name = "ssctl";

int send_vmap(__u32 localaddr, __u32 remoteaddr, int argc, char **argv)
{
  __u32 *ssdata;
  int i, ssdata_cnt;
  __u16 snapshotcnt;

  if (argc == 0) {
    fprintf(stderr, "Invalid parameters\n");
    return -1;
  }

  snapshotcnt = (__u16) atoi(argv[0]);

  argc--;
  argv++;

  ssdata_cnt = argc;
  ssdata = malloc(ssdata_cnt * sizeof(__u32));
  if (ssdata == NULL)
    return -1;

  for (i = 0; i < ssdata_cnt; i++) {
    ssdata[i] = inet_addr(argv[i]);
  }

  ssmsg_send_vmap(localaddr, remoteaddr, snapshotcnt, ssdata, ssdata_cnt);

  free(ssdata);

  return 0;
}

int send_rmap(__u32 localaddr, __u32 remoteaddr, int argc, char **argv)
{
  __u32 *ssdata;
  int i, ssdata_cnt;
  __u8 connected;

  if (argc == 0) {
    fprintf(stderr, "Invalid parameters\n");
    return -1;
  }

  connected = (__u8) atoi(argv[0]);

  argc--;
  argv++;

  ssdata_cnt = argc;
  ssdata = malloc(ssdata_cnt * sizeof(__u32));
  if (ssdata == NULL)
    return -1;

  for (i = 0; i < ssdata_cnt; i++) {
    ssdata[i] = inet_addr(argv[i]);
  }

  ssmsg_send_rmap(localaddr, remoteaddr, connected, ssdata, ssdata_cnt);

  free(ssdata);

  return 0;
}

int send_smap(__u32 localaddr, __u32 remoteaddr, int argc, char **argv)
{
  __u32 *ssdata;
  int i, ssdata_cnt;

  if (argc == 0) {
    fprintf(stderr, "Invalid parameters\n");
    return -1;
  }

  ssdata_cnt = argc;
  ssdata = malloc(ssdata_cnt * sizeof(__u32));
  if (ssdata == NULL)
    return -1;

  for (i = 0; i < ssdata_cnt; i++) {
    ssdata[i] = inet_addr(argv[i]);
  }

  ssmsg_send_smap(localaddr, remoteaddr, ssdata, ssdata_cnt);

  free(ssdata);

  return 0;
}

int send_snapshot(__u32 localaddr, __u32 remoteaddr, __u32 seqno,
                  int argc, char **argv)
{
  __u32 *ssdata;
  int i, ssdata_cnt;

  if (argc % 2 != 0) {
    fprintf(stderr, "Invalid parameters\n");
    return -1;
  }

  ssdata_cnt = argc;
  ssdata = malloc(ssdata_cnt * sizeof(__u32));
  if (ssdata == NULL)
    return -1;

  for (i = 0; i < ssdata_cnt; i += 2) {
    ssdata[i] = inet_addr(argv[i]);
    ssdata[i + 1] = htonl(atol(argv[i + 1]));
  }

  ssmsg_send_snapshot(localaddr, remoteaddr, seqno, ssdata, ssdata_cnt);

  free(ssdata);

  return 0;
}

void exit_print_usage(char **argv)
{
  printf("%s v%s\n\n"
         "Usage: %s src_ip dest_ip message [data]\n\n"
         "Message: x  - RESET               Reset the snapshot client\n"
         "         vm - VMAP                Load a VMAP into the snapshot client\n"
         "              data: VMAP data\n"
         "         rm - RMAP                Load a RMAP into the snapshot client\n"
         "              data: RMAP data\n"
         "         sm - SMAP                Load a SMAP into the snapshot client\n"
         "              data: SMAP data\n"
         "         c  - CONTROL             Send a control to the snapshot client\n"
         "              data: seqno\n"
         "         m  - MARKER              Send a marker to the snapshot client\n"
         "              data: seqno\n"
         "         s  - SNAPSHOT            Send a snapshot to the snapshot client\n"
         "              data: SNAPSHOT message\n"
         "         z  - CLEAR               Clear the data in the snapshot client\n\n",
         program_name, program_version, program_name);
  exit(1);
}

int main(int argc, char **argv)
{
  __u32 localaddr;
  __u32 remoteaddr;
  __u32 seqno;

  if (argc < 4)
    exit_print_usage(argv);

  localaddr = inet_addr(argv[1]);
  remoteaddr = inet_addr(argv[2]);

  if (strcmp(argv[3], "x") == 0) {
    // RESET
    if (argc != 4)
      exit_print_usage(argv);
    ssmsg_send_reset(localaddr, remoteaddr);
  } else if (strcmp(argv[3], "vm") == 0) {
    // VMAP
    if (argc < 4)
      exit_print_usage(argv);
    send_vmap(localaddr, remoteaddr, argc - 4, argv + 4);
  } else if (strcmp(argv[3], "rm") == 0) {
    // RMAP
    if (argc < 4)
      exit_print_usage(argv);
    send_rmap(localaddr, remoteaddr, argc - 4, argv + 4);
  } else if (strcmp(argv[3], "sm") == 0) {
    // SMAP
    if (argc < 4)
      exit_print_usage(argv);
    send_smap(localaddr, remoteaddr, argc - 4, argv + 4);
  } else if (strcmp(argv[3], "c") == 0) {
    // CONTROL
    if (argc != 5)
      exit_print_usage(argv);
    seqno = atol(argv[4]);
    ssmsg_send_control(localaddr, remoteaddr, seqno);
  } else if (strcmp(argv[3], "m") == 0) {
    // MARKER
    if (argc != 5)
      exit_print_usage(argv);
    seqno = atol(argv[4]);
    ssmsg_send_marker(localaddr, remoteaddr, seqno);
  } else if (strcmp(argv[3], "s") == 0) {
    // SNAPSHOT
    if (argc < 5)
      exit_print_usage(argv);
    seqno = atol(argv[4]);
    send_snapshot(localaddr, remoteaddr, seqno, argc - 5, argv + 5);
  } else if (strcmp(argv[3], "z") == 0) {
    // CLEAR
    if (argc != 4)
      exit_print_usage(argv);
    ssmsg_send_clear(localaddr, remoteaddr);
  } else {
    printf("Invalid option.\n");
  }

  return 0;
}
