#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <router/ddos/traceback/traceback.h>

#define MAX_CNT 10000
#define STAT_SIZE 10
#define NODECNT_SIZE 100

int stat[STAT_SIZE][NODECNT_SIZE];
int nodecnt[NODECNT_SIZE];

static void printFreq(const MarkingFieldsNode* fieldList, int statidx)
{
  const MarkingFieldsNode* ptr;
  int min, i = 0;
  int ok;

  if (fieldList == NULL)
    return;

  min = INT_MAX;
  
  for (ptr = fieldList->next; ptr != NULL; ptr = ptr->next) {
    ok = (min >= ptr->frequency);
    if (ok)
      stat[statidx][i]++;

    printf("%ld,%d,%d\t",
           ptr->markingField.startIP,
           ptr->frequency,
	   ok);

    min = ptr->frequency;
    i++;
  }

  printf("\n");
}

static void construct(MarkingFieldsNode *fieldList,
                      MarkingFields *fields,
                      int cnt, double elapsed, int statidx)
{
  RouterTree* head = newRouterTree(0, 10);
  int i;

  printf("%lf\t", elapsed);

  //  1. add all the fields into a MarkingFieldNode structure
  for (i=0; i<cnt; i++) {
    addMarkingFields(fieldList, &fields[i]);
  }

  //  2. Construct attack Graph, stored in a RouterTree structure
  attackGraphCon(head, fieldList);

  printFreq(fieldList, statidx);
  
  freeRouterTree(head);
}

/*
static int stat2(double elapsed) {
  int min;
  int ok = 1;
  int i;
  
  min = INT_MAX;

  printf("%lf\t", elapsed);

  for (i=1; i<11; i++) {
    printf("%d,%d\t", i, nodecnt[i]);
    if (nodecnt[i] > min)
      ok = 0;
    min = nodecnt[i];
  }
  printf("%d\n", ok);

  return ok;
}
*/

int minstat(int statidx, int nnode)
{
  int i, min;

  min=stat[statidx][1];

  for (i=2; i<nnode; i++) {
    if (stat[statidx][i] < min)
      min = stat[statidx][i];
  }
  return min;
}

int main(int argc, char **argv)
{
  MarkingFieldsNode *fieldList = NULL;
  MarkingFields fields[MAX_CNT];
  unsigned long long starttime = 0, elapsed = 0;
  double t;
  int i, j, cnt = 0;
  int ret, state;
  int nsample = 0;
 
  int param_nnode;
  double param_timeunit;
  int param_ntimeunit;
  int param_nsample;

  if (argc != 5) {
    printf("Usage: %s <num of nodes> <time unit> <num of time units> <num of samples>\n", argv[0]);
    exit(1);
  }
  
  param_nnode = atoi(argv[1]);
  param_timeunit = atof(argv[2]);
  param_ntimeunit = atoi(argv[3]);
  param_nsample = atoi(argv[4]);

  state = 1;

  printf("SAMPLE %d\n", nsample+1);


  for (j=0; j<NODECNT_SIZE; j++) {
    for (i=0; i<STAT_SIZE; i++) {
      stat[i][j] = 0;
    }
    nodecnt[j] = 0;
  }
  
  fieldList = newMarkingFieldsNode();
  do {
    ret = scanf("%llu\t%lu\t%lu\t%lu\n",
                &elapsed,
                &fields[cnt].startIP,
                &fields[cnt].endIP,         
                &fields[cnt].distance);

    if (ret == EOF)
      break;
    
    /*
    printf("%llu\t%lu\t%lu\t%lu\n",
           elapsed,
           fields[cnt].startIP,
           fields[cnt].endIP,         
           fields[cnt].distance);
    */
    cnt++;
    if (cnt >= MAX_CNT) {
      fprintf(stderr, "Error: field array overflow (%d >= %d)\n",
              cnt, MAX_CNT);
      exit(1);
    }

    nodecnt[fields[cnt].startIP]++;
    
    if (state == 0) {
      printf("%lf\n", elapsed/1000000.0);
      starttime = elapsed;
      state = 1;
    }
    
    elapsed -= starttime;
    t = elapsed/1000000.0;

    if (elapsed >= state * param_timeunit * 1000000) {
      construct(fieldList, fields, cnt, t, state-1);
      
      state++;
      cnt = 0;
      
      if (state > param_ntimeunit) {
	// analysis current sample
        nsample++;
        for (i=0; i<param_ntimeunit; i++) {
          printf("%f\t", ((double)minstat(i, param_nnode)) / nsample);
        }
        printf("\n");

	if (nsample >= param_nsample)
	  break;

	// begin next sample
        state = 0;
        printf("@\n");
        printf("SAMPLE %d\n", nsample+1);
        if (fieldList != NULL) {
          freeMarkingFieldsNode(fieldList);
        }
        fieldList = newMarkingFieldsNode();

        for (i=0; i<NODECNT_SIZE; i++) {
          nodecnt[i] = 0;
        }

      }
    }
  } while (1);

  if (fieldList != NULL) {
    freeMarkingFieldsNode(fieldList);
  }
  
  return 0;
}
