#include <arpa/inet.h>
#include <router/ddos/traceback/attackgraph.h>

/*
 *  MarkingFieldsNode functions:
 */

//////////////////////////////////////////////////

MarkingFieldsNode* newMarkingFieldsNode()
{
  MarkingFieldsNode* n = (MarkingFieldsNode*)malloc(sizeof(MarkingFieldsNode));
  n->next = NULL;
  return n;
}

void freeMarkingFieldsNode(MarkingFieldsNode* head)
{
  MarkingFieldsNode* ptr, *temp;
  for (ptr=head; ptr!=NULL; )
  {
    temp = ptr->next;
    free(ptr);
    ptr = temp;
  }
}

MarkingFieldsNode* dupMarkingFieldsNode(const MarkingFieldsNode* node)
{
  MarkingFieldsNode* copy;
  if (node == NULL)
    return NULL;
  copy = newMarkingFieldsNode();
  copy->markingField = node->markingField;
  copy->frequency = node->frequency;
  return copy;
}

MarkingFieldsNode* dupMarkingFieldsNodeList(const MarkingFieldsNode* list)
{
  MarkingFieldsNode* copy;
  if (list == NULL)
    return NULL;
  copy = dupMarkingFieldsNode(list);
  copy->next = dupMarkingFieldsNodeList(list->next);
  return copy;
}

MarkingFieldsNode* findMarkingFieldsNode(const MarkingFieldsNode* head,
                                         const MarkingFields fields)
{
  MarkingFieldsNode* ptr;
  for (ptr=head->next; ptr!= NULL; ptr=ptr->next)
  {
    if (ptr->markingField.startIP == fields.startIP &&
        ptr->markingField.endIP == fields.endIP &&
        ptr->markingField.distance == fields.distance)
      return ptr;
  }
  return NULL;
}
/* DEBUG function */
int printGraphList(Router* head)
{
  Router* ptr;
  Router* ptr2;
  struct in_addr in;
  
  if (head == NULL)
  {
    printf("(NULL)\n");
    return 0;
  }
  ptr = head->nextRouter;
  while (1)
  {
    //printf("print head: %d\n", head->ip);
    if (ptr == NULL)
      break;
    in.s_addr = htonl(ptr->ip);
    printf("%s(%lf) -> ", inet_ntoa(in), ptr->localTrafficRate);
    ptr2 = ptr->nextConnection;
    while (1)
    {
      if (ptr2 == NULL)
        break;
      in.s_addr = htonl(ptr2->ip);
      printf("%s(%lf), ", inet_ntoa(in), ptr2->localTrafficRate);
      ptr2 = ptr2->nextConnection;
    }
    printf("\n");
    ptr = ptr->nextRouter;
  }
  printf("\n");
  return 1;
}
////////////////////////////////////


Router* dupRouter(Router* ptr)
{
  Router* newPtr;
  newPtr = (Router*)malloc(sizeof(Router));
  newPtr->ip = ptr->ip;
  newPtr->nextRouter = NULL;
  newPtr->nextConnection = NULL;
  return newPtr;
}

/* insert Router (check dupicate) */
int insertRouter(Router* head, Router* pt)
{
  Router* ptr;
  Router* backPtr = NULL;
  ptr = head;
  while (ptr != NULL)
  {
    if (ptr->ip == pt->ip)
      return 0;
    backPtr = ptr;
    ptr = ptr->nextRouter;
  }
  /*printf("Insert: %d\n", pt->ip);*/
  backPtr->nextRouter = pt;
  pt->nextRouter = NULL;
  return 1;
}

/* insert Connection */
int insertConnection(Router* head, Router* from, Router* to)
{
  Router* ptr, *ptr2, *backPtr2;
  ptr = head;
  while (ptr != NULL)
  {
    if (ptr->ip == from->ip)
    {
      backPtr2 = ptr;
      ptr2 = ptr->nextConnection;
      while (ptr2 != NULL)
      {
        if (ptr2->ip == to->ip)
          return 0;
        backPtr2 = ptr2;
        ptr2 = ptr2->nextConnection;
      }
      /*printf("InsertConnection: %d -> %d\n", from->ip, to->ip);
       * */
      backPtr2->nextConnection = to;
      to->nextConnection = NULL;
      return 1;
    }
        
    ptr = ptr->nextRouter;
  }
  /* if finially cannot find the point, insert it and insert the 
   * connection again */
  insertRouter(head, from);
  return insertConnection(head, from, to);
}


int haveRouter(Router* head, Router* pt)
{
  Router* ptr;

  ptr = head;
  while (ptr != NULL)
  {
    if (ptr->ip == pt->ip)
      return 1;
    ptr = ptr->nextRouter;
  }
  return 0;
}

void freeList(Router* head)
{
  Router *ptr, *ptr2;
  Router *ptrN, *ptr2N;

  ptr = head;
  while (ptr != NULL)
  {
    ptrN = ptr->nextRouter;
    ptr2 = ptr->nextConnection;
    free(ptr);
    while (ptr2 != NULL)
    {
      ptr2N = ptr2->nextConnection;
      free(ptr2);
      ptr2 = ptr2N;
    }
    ptr = ptrN;
  }
}
////////////////////////////////////////////



int isMarkingFieldEqual(MarkingFields* a, MarkingFields* b)
{
  if (a->startIP == b->startIP &&
      a->endIP == b->endIP &&
      a->distance == b->distance)
    return 1;
  else
    return 0;
}

/*
 *  addMarkingFields
 *
 *  
 */
int addMarkingFields(MarkingFieldsNode* head, MarkingFields* field)
{
  MarkingFieldsNode* ptr, *oldptr;
  oldptr = head;
  for (ptr=head->next; ptr!=NULL; ptr=ptr->next)
  {
    if (field->distance < ptr->markingField.distance)
      break;
    if (isMarkingFieldEqual(&(ptr->markingField), field))
    {
      ptr->frequency++;
      return 1;
    }
    oldptr = ptr;
  }
  oldptr->next = newMarkingFieldsNode();
  memcpy(&(oldptr->next->markingField), field, sizeof(MarkingFieldsNode));
  oldptr->next->frequency = 1;
  oldptr->next->next = ptr;
  return 1;
}



/*
 *  uniqueFields
 *
 *  It returns a MarkingFieldsNode link list that contain only unique 
 *  MarkingFields, and it is sorted by distance in asc order.
 *
 */
/*
  MarkingFieldsNode* uniqueFields(MarkingFields* fieldArray, int sizeOfArray)
  {
  int i;
  int numOfNode = 0;
  int existed;
  MarkingFieldsNode* ptr;
  MarkingFieldsNode* oldPtr;
  MarkingFieldsNode* markingFieldsList = (MarkingFieldsNode*)malloc
  (sizeof(MarkingFieldsNode));
  markingFieldsList->next = NULL;
    
  for (i=0; i<sizeOfArray; i++)
  {
  existed = 0;
  oldPtr = markingFieldsList;
  for (ptr=markingFieldsList->next; ptr != NULL; ptr = ptr->next)
  {
  if (fieldArray[i].distance > ptr->markingField.distance)
  break;
  if (isMarkingFieldEqual(&fieldArray[i], &(ptr->markingField)))
  {
  existed = 1;
  break;
  }
  oldPtr = ptr;
  }
  // if the fieldArray[i] doesn't exist in uniqueArray...
  if (existed != 1)
  {
  oldPtr->next = (MarkingFieldsNode*)malloc(sizeof(MarkingFieldsNode));
  oldPtr->next->next = ptr;
  oldPtr->next->markingField = fieldArray[i];
  } 
  }
  return markingFieldsList;
  }
*/

/*
 * newRouter
 */
Router* newRouter()
{
  Router* r = (Router*)malloc(sizeof(Router));
  r->nextRouter = NULL;
  r->nextConnection = NULL;
  return r;
}



/*
 * RouterTree (Tree) functions:
 */
RouterTree* newRouterTree(int ip, int maxNumOfBranch)
{
  RouterTree* node = (RouterTree*)malloc(sizeof(RouterTree));
  node->ip = ip;
  node->numOfBranch = 0;
  node->maxNumOfBranch = maxNumOfBranch;
  node->nextLevelArray = (RouterTree**)calloc(maxNumOfBranch,
                                              sizeof(RouterTree*));
  node->localTrafficRate = 0;
  return node;
}
/*
  RouterTree* dupRouterTreeNode(const RouterTree* t)
  {
  if (t == NULL)
  return NULL;
  RouterTree* node = newRouterTree(t->ip, t->maxNumOfBranch);
  node->numOfBranch = 0;
  return node;
  }
*/
int insertTreeConnection(RouterTree* head, const MarkingFields* node)
{
  int i;
  if (head == NULL)
    return 0;
  if (node == NULL)
    return 0;
  if (node->endIP == head->ip)
  {
    for (i=0; i<head->numOfBranch; i++)
    {
      if (node->startIP == head->nextLevelArray[i]->ip)
        return 0;
    }
    if (head->numOfBranch == head->maxNumOfBranch)
    {
      head->maxNumOfBranch = head->maxNumOfBranch*2;
      head->nextLevelArray = (RouterTree**)realloc(head->nextLevelArray,
                                                   head->maxNumOfBranch*sizeof(RouterTree*));
    }
    head->nextLevelArray[head->numOfBranch] = newRouterTree(node->startIP,
                                                            10);
    (head->numOfBranch)++;
    //printf("insertTreeConnection: add \"%08lx->%08lx\" to %08lx\n", node->startIP, node->endIP, head->ip);
    return 1;
  }
  else
  {
    for (i=0; i<head->numOfBranch; i++)
    {
      if (insertTreeConnection(head->nextLevelArray[i], node) != 0)
        return 1;
    }
    //printf("insertTreeConnection: not add \"%08lx->%08lx\" to %08lx\n", node->startIP, node->endIP, head->ip);
    return 0;
  }
}

void printTree(const RouterTree* head, const char* prefix)
{
  char newPrefix[500];
  int i;
  struct in_addr in;
    
  if (prefix != NULL)
    strncpy(newPrefix, prefix, 499);
  else
    newPrefix[0] = 0;

  in.s_addr = htonl(head->ip);
  printf("%s%s(%lf)\n", newPrefix, inet_ntoa(in), head->localTrafficRate);
  strncat(newPrefix, "|  ", 499);
  for (i=0; i<head->numOfBranch; i++)
  {
    printTree(head->nextLevelArray[i], newPrefix);
  }
}

void freeRouterTree(RouterTree* head)
{
  int i;
  for (i=0; i<head->numOfBranch; i++)
    freeRouterTree(head->nextLevelArray[i]);
  free(head->nextLevelArray);
  free(head);
}
/*
 *  attackGraphCon
 *
 *  It constructs the attack graph by a list of packet marking field (may have
 *  dup records).
 *
 *  head: head pointer of the graph
 *
 */
/*
  int attackGraphCon(Router* head, MarkingFieldsNode* markingFieldList)
  {
  MarkingFieldsNode* ptr;
  Router* newRouter, *newRouter2;

//    markingFieldList = uniqueFields(fieldArray, sizeOfArray);
for (ptr=markingFieldList->next; ptr!=NULL; ptr=ptr->next)
{
printf("%x, %x, %d\n", ptr->markingField.startIP, 
ptr->markingField.endIP, 
ptr->markingField.distance);
}
for (ptr=markingFieldList->next; ptr!=NULL; ptr=ptr->next)
{
newRouter = (Router*)malloc(sizeof(Router));
newRouter->ip = ptr->markingField.endIP;
newRouter2 = (Router*)malloc(sizeof(Router));
newRouter2->ip = ptr->markingField.startIP;
insertConnection(head, newRouter, newRouter2);
}
}
*/
int attackGraphCon(RouterTree* head, const MarkingFieldsNode* markingFieldList)
{
  MarkingFieldsNode* ptr;

  for (ptr=markingFieldList->next; ptr!=NULL; ptr=ptr->next)
  {
    insertTreeConnection(head, &ptr->markingField);
  }
  return 1;
}



/*
 *  attackPathCon
 *
 *  It is a recusive function.
 *  It constructs a set of attack path from the attack graph.
 *  return the following structure:
 *
 *  Router
 *  
 *  head
 *  |
 *  v
 *  (Host/router) -> (lvl 1 router A) -> (lvl 2 router A) -> NULL
 *  |
 *  v
 *  (Host/router) -> (lvl 1 router A) -> (lvl 2 router B) -> NULL
 *  |
 *  :
 *  |
 *  NULL
 *
 */
Router* attackPathCon(const RouterTree* head)
{
  int i, j;
  Router* ptr;
  Router* returnFromNextLevel, *ptr2;
  Router* headNode = newRouter();

  if (head == NULL)
    return NULL;

  headNode->nextRouter = NULL;
  headNode->nextConnection = NULL;
  ptr = headNode;
  if (head->numOfBranch == 0)
  {
    ptr->nextRouter = newRouter();
    ptr = ptr->nextRouter;
    ptr->ip = head->ip;
    ptr->localTrafficRate = head->localTrafficRate;
    ptr->nextConnection = NULL;
    ptr->nextRouter = NULL;
  }
  else
  {
    for (i=0; i<head->numOfBranch; i++)
    {
      returnFromNextLevel = attackPathCon(head->nextLevelArray[i]);
      // merge all the returned informatin into my struct
      j = 0;
      for (ptr2=returnFromNextLevel->nextRouter; ptr2!=NULL; ptr2=ptr2->nextRouter)
      {
        ptr->nextRouter = newRouter();
        ptr = ptr->nextRouter;
        ptr->ip = head->ip;
        ptr->localTrafficRate = head->localTrafficRate;
        ptr->nextConnection = ptr2;
        ptr->nextRouter = NULL;
        j++;
      }
    }
  }
  return headNode;
}

/*
 * extractPacketFrequency
 *
 * It extracts the packet frequency from the MarkingFieldsNode list to a int
 * array. The data in the int array is arranged by the attackPath, from
 * nearest to farest.
 *
 */
int* extractPacketFrequency(const MarkingFieldsNode* head, 
                            const Router* attackPath,
                            int* sizeOfArray)
{
  MarkingFieldsNode* nodePtr;
  MarkingFields temp;
  const Router* routerPtr;   
  int distance, i;
  int* returnArray;
    
  *sizeOfArray = 0;
  for (routerPtr=attackPath; routerPtr->nextConnection!=NULL; 
       routerPtr=routerPtr->nextConnection)
  {
    (*sizeOfArray)++;
  }
  returnArray = (int*)calloc(*sizeOfArray, sizeof(int));
    
  for (routerPtr=attackPath, distance=1, i=0; 
       routerPtr->nextConnection!=NULL; 
       routerPtr=routerPtr->nextConnection, distance++, i++)
  {
    temp.startIP = routerPtr->nextConnection->ip;
    temp.endIP = routerPtr->ip;
    temp.distance = distance;
        
    nodePtr = findMarkingFieldsNode(head, temp);
    if (nodePtr == NULL)
    {
      printf("extractPacketFrequency: Error: can't find the packet count of a node in the a attack path.\n");
      returnArray[i] = 0;
    }
    else
    {
      returnArray[i] = nodePtr->frequency;
    }
  }
  return returnArray;
    
}

void printMarkingFieldsNodeList(const MarkingFieldsNode* list)
{
  const MarkingFieldsNode* ptr;
  struct in_addr in;
  for (ptr=list; ptr!=NULL; ptr=ptr->next)
  {
//    in.s_addr = htonl(ptr->markingField.startIP);
    in.s_addr = ptr->markingField.startIP;
    printf("%s -> ", inet_ntoa(in)); 
//    in.s_addr = htonl(ptr->markingField.endIP);
    in.s_addr = ptr->markingField.endIP;
    printf("%s (%lu), f:%d\n", inet_ntoa(in),
           ptr->markingField.distance, ptr->frequency);
  }

}











/* the following code is commented out */
#if 0

/*
  This file contain functions that generate upstream router IP information 
  for each software router in the attack-graph.
  These information will be sent to the routers for the use of Snapshot 
  DDoS detection algorithm.
*/

/*
  Router* genRouterList(const RouterTree* treeHead)
    
  This function generate the upstream router list of all the node in the
  RouterTree.

  return structure:

  host router1 --nextConnection--> upstreamRouter1.1 --nextConnection--> upstreamRouter1.2 -> NULL
  |
  nextRouter
  |
  v
  host router2 --nextConnection--> upstreamRouter1 --nextConnection--> upstreamRouter2 ...
  |
  .
  .
  v
  NULL

  look for printRouterUpstreamList() for example use.

  ***: PLS REMEMBER to FREE the return Router* struct (by freeRouterTree
  function) about use~!
*/
Router* genRouterList(const RouterTree* treeHead)
{
  int i;
  Router *head, *temp, *ptr, *ptr2;

  if (treeHead == NULL)
    return 0;
  head = newRouter();
  head->ip = treeHead->ip;
  ptr = head;
  ptr2 = head;
    
  printf(">>> %lu\n", head->ip);

  if (headOfHead != NULL)
  {
    ptr2->nextConnection = newRouter();
    ptr2->nextConnection->ip = headOfHead->ip;
    ptr2 = ptr2->nextConnection;
  }
    
  for (i=0; i<treeHead->numOfBranch; i++)
  {
    printf("%d\n", i);
    ptr2->nextConnection = newRouter();
    ptr2->nextConnection->ip = treeHead->nextLevelArray[i]->ip;
    ptr2 = ptr2->nextConnection;
    ptr->nextRouter = genRouterList(treeHead->nextLevelArray[i]);
    while (ptr->nextRouter != NULL) // move to the end of the returned link list
      ptr = ptr->nextRouter;
  }
  ptr->nextRouter = NULL;
  ptr2->nextConnection = NULL;
  return head;
    
}

int routerList2File(const Router* routerList, char* nodeFilePathName, char* linkFilePathName)
{
  FILE* nodefp, *linkfp;
  Router* ptr, *ptr2;
  unsigned long zero;

  zero = 0;

  nodefp = fopen(nodeFilePathName, "w");
  linkfp = fopen(linkFilePathName, "w");
  if (nodefp == NULL || linkfp == NULL)
    return 0;

  for (ptr=routerList; ptr != NULL; ptr=ptr->nextRouter)
  {
    fwrite(ptr->ip, sizeof(unsigned long), 1, nodefp);
    for (ptr2=ptr->nextConnection; ptr2!=NULL; ptr2=ptr2->nextConnection)
      fwrite(ptr2->ip, sizeof(unsigned long), 1, linkfp);
    fwrite(&zero, sizeof(unsigned long), 1, linkfp);
  }
  fclose(nodefp);
  fclose(linkfp);
  return 1;
}

/*
 * DEBUG function
 *
 */
void printRouterUpstreamList(const Router* head)
{
  const Router *ptr, *ptr2;

  for (ptr=head; ptr!=NULL; ptr=ptr->nextRouter)
  {
    printf("%lu -> ", ptr->ip);
    for (ptr2=ptr->nextConnection; ptr2!=NULL; ptr2=ptr2->nextConnection)
      printf("%lu, ", ptr2->ip);
    printf("\n");
  }
}
#endif


RouterTreeListNode* newRouterTreeListNode(const RouterTree* parent, const RouterTree* n)
{
  RouterTreeListNode* r;
  r = (RouterTreeListNode*)malloc(sizeof(RouterTreeListNode));
  r->treenode = n;
  r->parentNode = parent;
  r->next = NULL;
  return r;
}

void freeRouterTreeListNode(RouterTreeListNode* head)
{
    if (head->next != NULL)
        freeRouterTreeListNode(head->next);
    free(head);
}

/*
 * This file contain function for writing the router tree structure into a
 * binary file (can view as an big u32 array) & in a bracket string format
 *
 * Structure: (eg. [ ] : a 32 bit position)
 *
 * [v.ip][lvl1 router1][lvl1 router2]...[lvl1 routerX][lvl2 router1]...[lvlN router Xn][0xFFFFFFFF]
 * [V upstream router1][V upstream router2] ... [0xFFFFFFFF]
 * [lvl1 router1 parent router][lvl1 router1 upstream router1] ...
 * .
 * ... [lvlN-1 routerXn-1 upstream router][0xFFFFFFFF]
 * [lvlN router1 parent router][0xFFFFFFF]      // for lvlN routers
 * .
 * [lvlN routerN parent router][0xFFFFFFF]
 *
 * eg. for a tree like this:
 *
 * 1 - 2 - 3 - 4
 * 5 -/       /
 *           /
 * 6 - - - -/
 *
 * output:
 * 4 3 6 2 1 5 FFFFFFFF 3 6 FFFFFFFF 4 2 FFFFFFFF 4 FFFFFFFF 3 1 5 FFFFFFFF 2 FFFFFFFF 2 FFFFFFFF
 *
 * when you read the numberOfRouter+1 0xFFFFFFFF, that is the end of file. (Do
 * you know why? Ask John Lau if you dont know)
 * 
 * pls free the return array by yourself.
 */
unsigned long* writeTree(const RouterTree *root)
{
  int i;
  RouterTreeListNode* queueHead, *queuePtr;
  RouterTreeListNode* nextqueueHead, *nextqueuePtr;
  unsigned long *routerList;
  unsigned long *routerConnection;
  int routerListSize = 8;
  int routerListIndex = 0;
  int routerConnectionSize = 8;
  int routerConnectionIndex = 0;

  if (root == NULL)
    return NULL;
    
  routerList = (unsigned long*)malloc(routerListSize*sizeof(unsigned long));
  routerListIndex = 0;
  routerConnection = (unsigned long*)malloc(routerConnectionSize*sizeof(unsigned long));
  routerConnectionIndex = 0;
  queueHead = newRouterTreeListNode(NULL, NULL);
  queueHead->next = newRouterTreeListNode(NULL, root);
  while (queueHead->next != NULL)
  {
    nextqueueHead = newRouterTreeListNode(NULL, NULL);
    nextqueuePtr = nextqueueHead;
    for (queuePtr=queueHead->next; queuePtr != NULL; queuePtr=queuePtr->next)
    {
            
      if (routerConnectionIndex+3+queuePtr->treenode->numOfBranch >= routerConnectionSize)  
      {                                                  
        routerConnectionSize += 3+queuePtr->treenode->numOfBranch;                    
        routerConnectionSize *= 2;                    
        printf("Realloc: routerConnection size: %d, pointer: %p\n", routerConnectionSize*sizeof(unsigned long), routerConnection);
        routerConnection = (unsigned long*)realloc(routerConnection,
                                                   routerConnectionSize*sizeof(unsigned long));
        printf("Realloc: >routerConnection size: %d, pointer: %p\n", routerConnectionSize*sizeof(unsigned long), routerConnection);
      }                                                                   
            
      // copy the ip to the routerList
      //printf("routerListIndex: %d, ip: %X\n", routerListIndex, queuePtr->treenode->ip);
      routerList[routerListIndex++] = queuePtr->treenode->ip;
      if (routerListIndex == routerListSize)
      {
          routerListSize *= 2;
        printf("Realloc: routerList size: %d, pointer: %p\n", routerListSize*sizeof(unsigned long), routerList);
          routerList = (unsigned long*)realloc(routerList, routerListSize*sizeof(unsigned long));
        printf("Realloc: >routerList size: %d, pointer: %p\n", routerListSize*sizeof(unsigned long), routerList);
      }
      // copy the upstream router information & enqueue the next level
      // routers
            
      // copy the parent router IP 1st.
      if (queuePtr->parentNode != NULL)
      {
        printf("routerConnection[%d] = %lX\n", routerConnectionIndex, queuePtr->parentNode->ip);
        routerConnection[routerConnectionIndex++] = queuePtr->parentNode->ip;
      }
      for (i=0; i<queuePtr->treenode->numOfBranch; i++)
      {
        // copy upstream router information to routerConnection[]
        printf("routerConnection[%d] = %lX\n", routerConnectionIndex, queuePtr->treenode->nextLevelArray[i]->ip);
        routerConnection[routerConnectionIndex++] = queuePtr->treenode->nextLevelArray[i]->ip;
        // enqueue the next level routers
        //printf("enq: ip=%x\n", queuePtr->treenode->nextLevelArray[i]->ip);
        nextqueuePtr->next = newRouterTreeListNode(queuePtr->treenode, queuePtr->treenode->nextLevelArray[i]);
        nextqueuePtr = nextqueuePtr->next;
      } 
      //printf("routerConnection[%d] = %X\n", routerConnectionIndex, 0xFFFFFFFF);
      routerConnection[routerConnectionIndex++] = 0xFFFFFFFF;
    }
    freeRouterTreeListNode(queueHead);
    queueHead = nextqueueHead;
  }
  freeRouterTreeListNode(queueHead);
    
  // copy back the information to routerList[]
    printf("Realloc: final size: %d, pointer: %p\n", routerListIndex+1+routerConnectionIndex+1*sizeof(unsigned long), routerList);
  routerList = (unsigned long*)realloc(routerList, (routerListIndex+1+routerConnectionIndex+1)*
                                       sizeof(unsigned long));
    printf("Realloc: >final size: %d, pointer: %p\n", routerListIndex+1+routerConnectionIndex+1*sizeof(unsigned long), routerList);
  routerList[routerListIndex++] = 0xFFFFFFFF;
  memcpy(routerList+routerListIndex, routerConnection, 
         routerConnectionIndex*sizeof(unsigned long));
  return routerList;
}


/*
 *  writeTreeStr
 *
 *  It will return a ()-type of tree representation.
 *  such as 0.0.0.1(0.0.0.2(0.0.0.3()0.0.0.4()))
 *
 *  Pls free the return String by yourself
 */
char* writeTreeStr(const RouterTree *node)
{
  char* returnStr;
  int returnStrSize = 20;
  struct in_addr in;
  char* temp;
  int i;
    
  returnStr = (char*)malloc(returnStrSize*sizeof(char));
  if (node == NULL)
  {
    strcpy(returnStr, "()");
    return returnStr;
  }
  in.s_addr = node->ip;
  strcpy(returnStr, inet_ntoa(in));
  strcat(returnStr, "(");
  for (i=0; i<node->numOfBranch; i++)
  {
    temp = writeTreeStr(node->nextLevelArray[i]);
    if (strlen(temp) + strlen(returnStr) >= returnStrSize)
    {
      returnStrSize = (strlen(temp) + strlen(returnStr))*2;
      returnStr = (char*)realloc(returnStr, returnStrSize*sizeof(char));
    }
    strcat(returnStr, temp);
    free(temp);
  }
  strcat(returnStr, ")");
  return returnStr;
}
