#include <stdlib.h>
#include <limits.h>
#include <linux/netfilter.h>
#include <libipq.h>
#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <linux/ip.h>
#include <math.h>
#include <router/ddos/traceback/traceback.h>

#ifdef __DEBUG__
#define DEBUGP(fmt, args...)                                            \
  printf("[%s:%s:%d]  " fmt "\n", __FILE__, __FUNCTION__, __LINE__, ##args)
#else
#define DEBUGP(fmt, args...) {}
#endif

#define BUFSIZE 2048
#define MAX_CNT 100

#define NIPQUAD(addr)          \
  ((unsigned char *)&addr)[0], \
  ((unsigned char *)&addr)[1], \
  ((unsigned char *)&addr)[2], \
  ((unsigned char *)&addr)[3]

static void die(struct ipq_handle *h)
{
  ipq_perror("passer");
  ipq_destroy_handle(h);
  exit(1);
}

int setkbufsize(struct ipq_handle *h, unsigned long rsize,
                unsigned long wsize)
{
  unsigned long bufsize, l;
  int ret;

  /* set kernel receive buffer size */
  DEBUGP("initializing kernel RCVBUF size ...\n");

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(h->fd, SOL_SOCKET, SO_RCVBUF,
                   &bufsize, (socklen_t *) & l);
  DEBUGP("current RCVBUF size is %lu\n", bufsize);

  bufsize = rsize;
  l = (socklen_t) sizeof(bufsize);
  ret = setsockopt(h->fd, SOL_SOCKET, SO_RCVBUF, &bufsize, l);
  DEBUGP("changing RCVBUF size to %lu\n", bufsize);

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(h->fd, SOL_SOCKET, SO_RCVBUF,
                   &bufsize, (socklen_t *) & l);
  DEBUGP("new RCVBUF size is %lu\n", bufsize);


  /* set kernel send buffer size */
  DEBUGP("initializing kernel RCVBUF size ...\n");

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(h->fd, SOL_SOCKET, SO_SNDBUF,
                   &bufsize, (socklen_t *) & l);
  DEBUGP("current SNDBUF size is %lu\n", bufsize);

  bufsize = wsize;
  l = (socklen_t) sizeof(bufsize);
  ret = setsockopt(h->fd, SOL_SOCKET, SO_SNDBUF, &bufsize, l);
  DEBUGP("changing SNDBUF size to %lu\n", bufsize);

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(h->fd, SOL_SOCKET, SO_SNDBUF,
                   &bufsize, (socklen_t *) & l);
  DEBUGP("new SNDBUF size is %lu\n", bufsize);

  return 0;
}

int main(int argc, char **argv)
{
  int status;
  unsigned char buf[BUFSIZE];
  struct ipq_handle *h;
  struct timeval starttime, endtime;
  int started = 0;
  unsigned long long elapsed = 0;

  h = ipq_create_handle(0, PF_INET);
  if (!h)
    die(h);

  status = ipq_set_mode(h, IPQ_COPY_PACKET, BUFSIZE);
  if (status < 0)
    die(h);

  setkbufsize(h, 65536 * 10, 65536 * 10);

  do {
    status = ipq_read(h, buf, BUFSIZE, 0);
    if (status < 0)
      die(h);

    switch (ipq_message_type(buf)) {
    case NLMSG_ERROR:
      fprintf(stderr, "Received error message %d\n", ipq_get_msgerr(buf));
      break;

    case IPQM_PACKET:
    {
      ipq_packet_msg_t *m = ipq_get_packet(buf);
      struct iphdr *iph = (struct iphdr *) m->payload;
      struct edgemark_opt *opt = EDGEMARK_OPTPTR(iph);

      if (((iph->ihl << 2) == EDGEMARK_IPHLEN) &&
          (opt->type == IPOPT_EDGEMARK) &&
          (opt->length == EDGEMARK_OPTLEN) &&
          (opt->msgtype == EDGEMARK_MSGTYPE_MARK))
      {
        if (started == 0) {
          started = 1;
          gettimeofday(&starttime, NULL);
        }

        gettimeofday(&endtime, NULL);

        elapsed =
          (unsigned long long) (endtime.tv_sec -
                                starttime.tv_sec) * 1000000 +
          (unsigned long long) (endtime.tv_usec - starttime.tv_usec);

        printf("%llu\t%d.%d.%d.%d\t%d.%d.%d.%d\t%d\n",
               elapsed,
               NIPQUAD(opt->mark_start),
               NIPQUAD(opt->mark_end), opt->mark_distance);

      }

      status = ipq_set_verdict(h, m->packet_id, NF_ACCEPT, 0, NULL);
      if (status < 0)
        die(h);
      break;
    }

    default:
      fprintf(stderr, "Unknown message type!\n");
      break;
    }

  } while (1);

  ipq_destroy_handle(h);
  return 0;
}
