// indicate this is an ipqext extension
#define _IPQEXT_EXTENSION_

#define __SYSLOG__
#define __DEBUG__

//----------------------------------------------------------------------//
//
//  ipqext extension module
//
//----------------------------------------------------------------------//

#include <linux/netfilter.h>
#include <libipq.h>
#include <router/ipqext/ipqext.h>
#include <lib/util.h>
#include <linux/ip.h>
#include <time.h>
#include <sys/file.h>
#include <stdlib.h>
#include "myArs.h"
#include "myArs.c"
#include "changeDetect.c"

/*
 * define extension name here
 *
 * extension name must match with the output module name
 * ( i.e. ipqext_{extension name} )
 */
const char extension_name[] = "dos";

//----------------------------------------------------------------------//
//
//  module functions
//

/*
 * ipqext structure
 */
static ipqext_t ext = {
  prev      : NULL,
  next      : NULL,
  handle    : NULL,
  active    : 0,
  name      : extension_name,
  version   : "1.0",
  groupmask : 0,
  init      : &init,
  fini      : &fini,
  proc      : &proc,
  start     : &start,
  stop      : &stop,
  reset     : &reset,
  info      : &info,
  help      : &help,
};

// 
//  Module variable and structure
// 

static ARS_info *ars_info;
static int      count;
static int      oldTime=0;
static int      startTime=0;
static double   samplingProb = 1;
static int      samplingSize = 100;
static int      samplingInterval = 1;   // HARD_CODE sampling interval = 1 sec
static double   mean;                   // reuseful varable for GUI display
// for change detection:
#define CALIBRATE_SAMPLES 50
static int      numOfBlockPassed = 0;
static ChangeDetectStruct* changeDetectS = NULL;
static CalibrateStruct* calibrateS = NULL;      
// calibrateStruct only malloc at the init, everytime it 'start' use the same
// calibrateStruct, no calibration again

/////////////////////////
//

/*
#define DATA_LOG "/tmp/ARS_data.log"
#define DETECTION_LOG "/tmp/ARS_detection.log"
#define DETECTION_PARA_h     "/tmp/_detection_temp_para_h.para"
#define DETECTION_PARA_v     "/tmp/_detection_temp_para_v.para"
#define CHANGE_DETECTED      "/tmp/_detection_temp_change"
*/

#define DATA_LOG "ARS_data.log"
#define DETECTION_LOG "ARS_detection.log"
#define DETECTION_PARA_h     "_detection_temp_para_h.para"
#define DETECTION_PARA_v     "_detection_temp_para_v.para"
#define CHANGE_DETECTED      "_detection_temp_change"

static void writeToFile(char* filename, char* string)
{
  FILE*   fin; 
  int     fd;
    
  fin = fopen(filename, "a");
  if (fin != NULL)
  {
    fd = fileno(fin);
    flock(fd, LOCK_EX);
        
    fprintf(fin, "%s", string);

    flock(fd, LOCK_UN);
    fclose(fin);
  }
}

static void flushFile(char* filename)
{
  FILE*   fin; 
  int     fd;
    
  fin = fopen(filename, "w");
  fd = fileno(fin);
  flock(fd, LOCK_EX);
    
  flock(fd, LOCK_UN);
  fclose(fin);
}

static int readFile(char* filename, double* returnNum)
{
  FILE*   fin; 
  int     fd;
  int     returnValue = 0;
  double  temp;
    
  fin = fopen(filename, "r");
  if (fin != NULL)
  {
    fd = fileno(fin);
    flock(fd, LOCK_EX);

    if (fscanf(fin, "%lf", &temp) > 0)
    {
      returnValue = 1;
      *returnNum = temp;
    }
        
    flock(fd, LOCK_UN);
    fclose(fin);
  }
  return returnValue;
}


/*
 * packet processing function
 *
 * return: NF_ACCEPT   packet is accepted and leaves ipqext immediately
 *         NF_QUEUE    packet is queued for further ipqext processing
 *         NF_DROP     packet is explicitly dropped
 *         NF_STOLEN   packet is stolen
 *
 */
int proc(ipq_packet_msg_t *m)
{
  unsigned int retval = NF_ACCEPT;
  
  int     newTime;
  int     size;
  char    line[1000];
  double  nowVolume;
  double  predictCount;
  double  g1t, g2t;
  struct  iphdr *iph;

  iph = (struct iphdr *) m->payload;
    
  newTime = m->timestamp_sec;
  size = m->data_len;

  if (newTime - oldTime >= samplingInterval)
  {
    ARS_finishBlock(ars_info, count, &samplingSize, &samplingProb);
        
    //mean = 0;         //we should reuse the last mean 
    predictCount = (double)samplingSize/samplingProb;
        
    // actcully, we should not access the data inside ars_info directly...
    if (ars_info->blockInfo != NULL)
    {
      if (ars_info->blockInfo->mean != 0)
        mean = ars_info->blockInfo->mean;
            
      //printf("count: %d\n", count);
      DB("t: %d\n", newTime);
      DB("mean: %lf, numOfPacket: %d\n", 
         ars_info->blockInfo->mean,
         ars_info->blockInfo->numOfPacket);
      DB("pSCV: %lf, pCount: %lf\n", 
         ars_info->predictSCV,
         ars_info->predictCount);
      DB("samplingSize: %d, samplingProb: %lf\n", 
         samplingSize, samplingProb);
    }

    sprintf(line, "%d\t%lf\t%d\t%lf\n", newTime-startTime, count*mean, count, mean);
    writeToFile(DATA_LOG, line);
    
    // Change detection
    nowVolume = mean * count;
    if (numOfBlockPassed < CALIBRATE_SAMPLES)           // calibrating state
    {
      AddCalibrateSample(calibrateS, nowVolume);
    }
    else if (numOfBlockPassed == CALIBRATE_SAMPLES)     // calibration
    {
      changeDetectS = Calibrate(calibrateS, 0, -1);
            
      //sprintf(line, "%d\t%lf\t%lf\t%lf\n", newTime-startTime, changeDetectS->h, g1t, g2t);
      //writeToFile(DETECTION_LOG, line);
            
      DB("Calibration Complete.\nStart change detection.\n");
      flushFile(DETECTION_LOG);
    }
    else                                                // change detection
    {
      if (ChangeDetect(changeDetectS, nowVolume, &g1t, &g2t) > 0)
      {
        DB("Change Detected at block %d\n", numOfBlockPassed);
        sprintf(line, "%d\n", newTime-startTime);
        writeToFile(CHANGE_DETECTED, line);
      }
      // output change detection information to a file
      // the output order should be sync with the GUI, 
      // 1:time, 2:a value use line, 3... : a value use impulues
      sprintf(line, "%d\t%lf\t%lf\t%lf\n", newTime-startTime, changeDetectS->h, g1t, g2t);
      writeToFile(DETECTION_LOG, line);

    }
    numOfBlockPassed++;

    // init for next block of sampling
    ARS_initBlock(ars_info, samplingSize);
    oldTime = newTime;
    count = 0;
  }
  else // within the same sampling interval, add into sample by random chance...
  {
    //printf("Add packet, size: %d\n", size);
    if ( ((double)random()/RAND_MAX) <= samplingProb)
    {
      ARS_addSample(ars_info, size);
    }
    count++;
  }

  
  return retval;
}

/*
 * extension start function
 *
 * return on success, -1 or nonzero value on error
 */
int start()
{
  startTime = time(NULL);
  flushFile(DATA_LOG);
  flushFile(DETECTION_LOG);
  flushFile(DETECTION_PARA_v);
  flushFile(DETECTION_PARA_h);
  flushFile(CHANGE_DETECTED);
  return 0;
}

/*
 * extension stop function
 *
 * return on success, -1 or nonzero value on error
 */
int stop()
{
  return 0;
}

/*
 * extension reset function
 *
 * return on success, -1 or nonzero value on error
 */
int reset()
{
  ARS_FreeARS_info(ars_info);
  ars_info = (ARS_info*)malloc(sizeof(ARS_info));
  ARS_initARS_info(ars_info, 5, 300, 0.1, 0.1);
  return 0;
}

/*
 * extension info function
 *
 * return a pointer to a char buffer or NULL
 */
const char infobuf[] = " info here ";

const char *info()
{
  return infobuf;
}

/*
 * extension help function
 *
 * return a pointer to a char buffer or NULL
 */
const char helpbuf[] = " help here ";

const char *help()
{
  DB("%s: help invoked\n", extension_name);
  return helpbuf;
}

/*
 *  initial pass function
 *
 * return on success, -1 or nonzero value on error
 */
int init_pass(void *handle, char *param, int paramlen)
{
  // set module handle
  ext.handle = handle;

  // parse parameter buffer

  return 0;
}

/*
 * extension initalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int init()
{
  // initalization codes here

  ars_info = (ARS_info*)malloc(sizeof(ARS_info));
  ARS_initARS_info(ars_info, 5, 300, 0.1, 0.1);
  calibrateS = NewCalibrateStruct(CALIBRATE_SAMPLES);

  count = 0;
  oldTime = 0;
  ARS_initBlock(ars_info, 1000);
  // register function must always be the last call
  ipqext_register(&ext);

  return 0;
}

/*
 * extension finalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int fini()
{
  // unregister function must always be the first call
  ipqext_unregister(&ext);

  // finialization codes here
  ARS_FreeARS_info(ars_info);
  FreeCalibrateStruct(calibrateS);

  return 0;
}

/*
 * module initalization function
 */
void _init(void)
{
}

/*
 * module finalization function
 */
void _fini(void)
{
}

//----------------------------------------------------------------------//
