
//////////////////////////
// Debug functions      //
//////////////////////////
void printARS_packetSizeArray(ARS_info *ars_info)
{
    int itemPerLine = 5;
    int i, j;
    printf("ars_info->numOfSample: %d\n",ars_info->numOfSample); 
    for (i=0; i<ars_info->numOfSample/itemPerLine+1; i++)
    {
        for (j=0; j<itemPerLine; j++)
            if (i*itemPerLine+j < ars_info->numOfSample)
                printf("%d\t", ars_info->packetSizeArray[i*itemPerLine+j]);
        printf("\n");
    }
}

void printARSError(ARS_info* ars_info)
{
    int i;
    BlockInfoNode* node;
    double scv_error;
    double count_error;
    
    node = ars_info->blockInfo;
    i = 0;
    while (1)
    {
        if (node == NULL)
            break;
        node = node->next;
        i++;
    }
    node = ars_info->blockInfo;
    while (1)
    {
        if (node == NULL)
            break;
        scv_error = node->pastPredictSCV - node->scv;
        count_error = node->pastPredictNumOfPacket - node->numOfPacket;
        //printf("%lf\n", scv_error);
        //printf("%lf\t%d\n", count_error, node->numOfPacket);
        printf("%d\t%d\t%lf\n", i, node->numOfPacket, node->pastPredictNumOfPacket);
        node = node->next;
        i--;
    }
}

//////////////////////////
// Internal functions   //
//////////////////////////

void freeAllBlock(ARS_info *ars_info)
{
    BlockInfoNode *nextptr;
    BlockInfoNode *ptr;
    ptr = ars_info->blockInfo;
    ars_info->blockInfo = NULL;
    while (ptr != NULL)
    {
        nextptr = ptr->next;
        free(ptr);
        ptr = nextptr;
    }
}



void freeLastBlock(ARS_info *ars_info)
{
    BlockInfoNode *preptr;
    BlockInfoNode *ptr;
    preptr = ars_info->blockInfo;
    for (ptr=ars_info->blockInfo; ptr!=NULL; ptr=ptr->next)
    {
        preptr = ptr;
    }
    if (preptr->next != NULL)
    {
        free(preptr->next);
        preptr->next = NULL;
    }
}


double calMean(ARS_info *ars_info)
{
    int i, totalSize;
    double mean;
    
    if (ars_info->numOfSample <= 0)
        return 0;
    totalSize = 0;
    for (i=0; i<ars_info->numOfSample; i++)
    {
        totalSize += ars_info->packetSizeArray[i];
    }
    mean = (double)totalSize / ars_info->numOfSample;
    return mean;
}


double calSCV(ARS_info *ars_info)
{
    int i, totalSize;
    double mean, standardD, scv;
    
    if (ars_info->numOfSample <= 0)
        return 0;
    totalSize = 0;
    for (i=0; i<ars_info->numOfSample; i++)
    {
        totalSize += ars_info->packetSizeArray[i];
    }
    mean = (double)totalSize / ars_info->numOfSample;
    standardD = 0;
    for (i=0; i<ars_info->numOfSample; i++)
    {
        standardD += pow((ars_info->packetSizeArray[i] - mean), 2);
    }
    standardD /= ars_info->numOfSample;
    scv = standardD/(mean*mean) ;
    return scv;
}

double calSCVCovariance(ARS_info *ars_info, int distance)
{
    int i;
    double total, covariance;
    BlockInfoNode* earlyNode;
    BlockInfoNode* lateNode;
        
    if (ars_info->blockInfo == NULL) 
        return 0;
    lateNode = ars_info->blockInfo;
    earlyNode = ars_info->blockInfo;
    for (i=0; i<distance; i++)
    {
        earlyNode = earlyNode->next;
        if (earlyNode == NULL)
            return 0;
    }
    total = 0;
    for (i=0; i<ars_info->numOfHistory; i++)
    {
        if (earlyNode == NULL)
            break;
        total += lateNode->scv * earlyNode->scv;
        lateNode = lateNode->next;
        earlyNode = earlyNode->next;
    }
    if (ars_info->numOfHistory == 0)
        printf("calSCVCovariance: numOfHistory == 0\n");
    covariance = total / i;
    return covariance;
}

double calCountCovariance(ARS_info *ars_info, int distance)
{
    int i;
    double total, covariance;
    BlockInfoNode* earlyNode;
    BlockInfoNode* lateNode;
        
    if (ars_info->blockInfo == NULL) 
        return 0;
    lateNode = ars_info->blockInfo;
    earlyNode = ars_info->blockInfo;
    for (i=0; i<distance; i++)
    {
        earlyNode = earlyNode->next;
        if (earlyNode == NULL)
            return 0;
    }
    total = 0;
    for (i=0; i<ars_info->numOfHistory; i++)
    {
        if (earlyNode == NULL)
            break;
        total += lateNode->numOfPacket * earlyNode->numOfPacket;
        lateNode = lateNode->next;
        earlyNode = earlyNode->next;
    }
    covariance = total / i;
    return covariance;
}

int calYuleWalkerMatrix(double* corelationArray, int size, 
                        double* coefficients)
{
    int i, j;
    double **matrix;
    //double *coef;

    matrix = (double **)malloc(sizeof(double *)*size);
    for (i=0;i<size;i++)
        matrix[i] = (double*)malloc(sizeof(double)*size);
    for (i=0; i<size; i++)
    {
        coefficients[i] = corelationArray[i+1];
        for (j=0; j<size; j++)
        {
            if (i == j)
                matrix[i][j] = 1;
            else if (i < j)
                matrix[i][j] = corelationArray[j-i];
            else if (i > j)
                matrix[i][j] = corelationArray[i-j];
        }
    }
    /*
    // Debug Display
    for (i=0; i<size; i++)
    {
        for (j=0; j<size; j++)
        {
            printf("%lf ", matrix[i][j]);
        }
        printf("\n");
    }
    for (i=0; i<size; i++)
        printf("co_in: %lf\n", coefficients[i]);
        */
    SolveLE(matrix, coefficients, size);
    
    //for (i=0; i<size; i++)
      //  printf("co_out: %lf\n", coefficients[i]);
    
    for (i=0;i<size;i++)
        free(matrix[i]);
    free(matrix);
    return 1;
}

/*
 *  Function calSCVModelParameters
 *
 *  This function calculate the Model Parameters for SCV.
 *
 *  Return Value: 0 if no error;
 *                1 if the 1st covariance is 0
 *           ***: The 1st covariance is 0 usually identical all the packets
 *                have the same size, it will make this ARS algorithm cannnot
 *                function. Because the algo depends on the variation of the 
 *                packet size.
 *                (The error usually happened in simple simulation case)
 */
int calSCVModelParameters(ARS_info *ars_info, int u, int v, double* coefficients)
{
    int i;
    double* covarianceArray;
    double* corelationArray;
    int returnValue;
    //FILE* fptr;
    
    returnValue = 1;
    covarianceArray = (double*)malloc((v+1)*sizeof(double));
    corelationArray = (double*)malloc((v+1)*sizeof(double));
    for (i=0; i<v+1; i++)
    {
        covarianceArray[i] = calSCVCovariance(ars_info, i);
        if (covarianceArray[0] != 0)
            corelationArray[i] = covarianceArray[i]/covarianceArray[0];
        else    // if all the packets have the same size, it will come to here.
        {
            corelationArray[i] = LONG_MAX;
            returnValue = 0;
        }
        //printf("covariance %d: %lf\n", i, covarianceArray[i]);
        //printf("corelation %d: %lf\n", i, corelationArray[i]);
    }
   
    //if (!AutoRegression(corelationArray,u+1,u,coefficients,LEASTSQUARES))
    //    fprintf(stderr,"AR routine failed\n");
    //for (i=0;i<u;i++)
    //    printf("%lf\n", coefficients[i]);
    calYuleWalkerMatrix(corelationArray, v, coefficients);
    
    free(covarianceArray);
    free(corelationArray);
    return returnValue;
}

int calCountModelParameters(ARS_info *ars_info, int u, int v, double* coefficients)
{
    int i;
    double* covarianceArray;
    double* corelationArray;
    //FILE* fptr;
    
    // The array should include Array[v]
    covarianceArray = (double*)malloc((v+1)*sizeof(double));
    corelationArray = (double*)malloc((v+1)*sizeof(double));
    for (i=0; i<v+1; i++)
    {
        covarianceArray[i] = calCountCovariance(ars_info, i);
        corelationArray[i] = covarianceArray[i]/covarianceArray[0];
        //printf("covariance %d: %lf\n", i, covarianceArray[i]);
        //printf("corelation %d: %lf\n", i, corelationArray[i]);
    }
   
    //if (!AutoRegression(corelationArray,u+1,u,coefficients,LEASTSQUARES))
    //    fprintf(stderr,"AR routine failed\n");
    //for (i=0;i<u;i++)
    //    printf("%lf\n", coefficients[i]);
    calYuleWalkerMatrix(corelationArray, v, coefficients);
    free(covarianceArray);
    free(corelationArray);
    return 1;
}

double calPredictSCV(ARS_info *ars_info, double* coefficients, int u)
{
    int i;
    double predictSCV;
    BlockInfoNode* node;

    node = ars_info->blockInfo;
    predictSCV = 0;
    for (i=0; i<u; i++)
    {
        if (node == NULL)
            break;
        predictSCV += coefficients[i]*(node->scv);
        node = node->next;
    }
    return predictSCV;
}

double calPredictCount(ARS_info *ars_info, double* coefficients, int u)
{
    int i;
    double predictCount;
    BlockInfoNode* node;

    node = ars_info->blockInfo;
    predictCount = 0;
    
    for (i=0; i<u; i++)
    {
        if (node == NULL)
            break;
        //printf("%lf * %d + ", coefficients[i], node->numOfPacket);
        predictCount += coefficients[i]*(node->numOfPacket);
        node = node->next;
    }
    //printf("\n");
    return predictCount;
}


double calZp(double relativeError, double probabilityOfError)
{
    return pow( (PointNormal(1-probabilityOfError/2) / relativeError) , 2);
}



int isPastTeamsZero(ARS_info* ars_info, int numOfTeams)
{
    int i;
    BlockInfoNode* ptr;

    ptr = ars_info->blockInfo;
    for (i=0; i<numOfTeams; i++)
    {
        if (ptr == NULL)
            break;
        //printf(">>>> %lf \n", ptr->mean);
        if (ptr->mean > 0.5)
            break;
        ptr = ptr->next;
    }
    if (i==numOfTeams)
        return 1;
    else
        return 0;
}




//////////////////////////


//////////////////////////
// Exteral function     //
//////////////////////////
/*
 *  function for a block of data
 */
int ARS_initBlock(ARS_info *ars_info, int packetSizeArraySize)
{
    if (ars_info->packetSizeArray != NULL)
        free(ars_info->packetSizeArray);
    if (packetSizeArraySize < 0)
        packetSizeArraySize = 0;
    ars_info->packetSizeArraySize = packetSizeArraySize;
    ars_info->packetSizeArray = (int*)malloc(packetSizeArraySize*sizeof(int));
    ars_info->numOfSample = 0;
    if (ars_info->packetSizeArray != NULL)
        return 1;
    else
        return 0;
}

int ARS_addSample(ARS_info *ars_info, int packetSize)
{
    if (ars_info->packetSizeArray == NULL ||
        ars_info->numOfSample >= ars_info->packetSizeArraySize)
        return 0;
    ars_info->packetSizeArray[ars_info->numOfSample] = packetSize;
    ars_info->numOfSample++;
    return 1;
}

int ARS_finishBlock(ARS_info *ars_info, int numOfPacket, 
                    int* predictSamplingCount, double* predictSamplingProb)
{
    int returnValue;
    double temp;
    BlockInfoNode* newNode;
    double* coefficients;
    double predictSCV;
    double predictCount; 
    double Zp;

    double SCVError, countError;

    newNode = (BlockInfoNode*)malloc(sizeof(BlockInfoNode));
    newNode->scv = calSCV(ars_info);
    newNode->numOfPacket = numOfPacket;
    newNode->mean = calMean(ars_info);
    newNode->pastPredictSCV = 0;
    newNode->pastPredictNumOfPacket = 0;
    //printf("real SCV:\t%lf\t\treal count:\t%d\n", newNode->scv, newNode->numOfPacket);
    newNode->next = ars_info->blockInfo;
    ars_info->blockInfo = newNode;
    
    ars_info->numOfHistory++;
    
    //printf("%lf\n", newNode->scv);
    
    if (ars_info->numOfHistory > ars_info->V)
    {
        // calculate the Error between predict result and real result
        if (ars_info->numOfHistory > ars_info->V+1)
        {
            newNode->pastPredictSCV = ars_info->predictSCV;
            newNode->pastPredictNumOfPacket = ars_info->predictCount;
            
            SCVError = ars_info->predictSCV - newNode->scv;
            countError = ars_info->predictCount - newNode->numOfPacket;
            //printf("%lf\t\t%lf\n", SCVError, countError);
            //printf("%lf\n", countError);
        
            if (ars_info->numOfHistory > ars_info->U)
            {
                // Free the last block of history, it is no use
                freeLastBlock(ars_info);
                ars_info->numOfHistory--;
            }
        }
        
        // Predict the next block data
        coefficients = (double*)malloc((ars_info->V)*sizeof(double));
        returnValue = calSCVModelParameters(ars_info, ars_info->V, 
                                            ars_info->V, coefficients);

        predictSCV = calPredictSCV(ars_info, coefficients, ars_info->V);
        ars_info->predictSCV = predictSCV;
        
        calCountModelParameters(ars_info, ars_info->V, ars_info->V, coefficients);
        predictCount = calPredictCount(ars_info, coefficients, ars_info->V);
        ars_info->predictCount = predictCount;
        //printf("Predict SCV:\t%lf\t\tPredict Count:\t%lf\n", predictSCV, predictCount);

        // Calculate the Zp in each round 
        // (May make is constant if never change it dynamically)
        Zp = calZp(ars_info->relativeError, ars_info->probabilityOfError);
        temp = (double)Zp * predictSCV;
        //printf("Next Sampling Count: %lf\n", temp);
        *predictSamplingCount = (int)temp;
        *predictSamplingProb = temp / predictCount;

        if (*predictSamplingCount <= 0)
        {
#ifdef DEBUG
            //printf("myARS: Warning: Prediction error due to same packet size ");
            //printf("of all the packets in a block.\n");
#endif
            //printf("myARS: Warning: Prediction error due to same packet size of all the packets in a block.\n");
            // if the past 3 turns are 0 too, reset the sampling parameter
            if (isPastTeamsZero(ars_info, 3))
            {
                freeAllBlock(ars_info);
                ars_info->numOfHistory=0;
                
                printf("Reset sampling para\n");
                *predictSamplingCount = 1000;
                *predictSamplingProb = 1;
            }
        }
    }
    else
    {
        // Default sampling parameter
        *predictSamplingCount = 1000;
        *predictSamplingProb = 1;
    }
    
    return 1;
}

int ARS_initARS_info(ARS_info* info, int V, int U, double relativeError, 
                 double probabilityOfError)
{
    if (V <= 0)
        return 0;
    if (relativeError <= 0)
        return 0;
    if (probabilityOfError <= 0)
        return 0;
    info->packetSizeArray = NULL;
    info->packetSizeArraySize = 0;
    info->numOfSample = 0;
    info->blockInfo = NULL;
    info->numOfHistory = 0;
    info->V = V;
    info->U = U;
    info->relativeError = relativeError;
    info->probabilityOfError = probabilityOfError;
    info->predictSCV = 0;
    info->predictCount = 0;
    return 1;
}


void ARS_FreeARS_info(ARS_info* info)
{
    BlockInfoNode* ptr;
    BlockInfoNode* ptr_next;
    
    free(info->packetSizeArray);

    ptr = info->blockInfo;
    while (1)
    {
        if (ptr == NULL)
            break;
        ptr_next = ptr->next;
        free(ptr);
        ptr = ptr_next;
    }
}
