#define __DEBUG__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>

#include <lib/util.h>
#include <router/ipqext/rsp.h>
#include <router/ipqext/extensions/ipqext_fec/fec.h>
#include <router/ipqext/extensions/ipqext_fec/encqueue.h>
#include <router/ipqext/extensions/ipqext_fec/ipqext_fec.h>

static inline void eq_free_slot(void **slot)
{
  if ((*slot) != NULL) {
    free(((unsigned char *) (*slot)) - sizeof(struct fechdr));
    (*slot) = NULL;
  }
}

static inline void eq_alloc_slot(void **slot, int size)
{
  eq_free_slot(slot);
  (*slot) = malloc(size + sizeof(struct fechdr));
  (*slot) += sizeof(struct fechdr);
}

inline void *eq_get_slot_head(void *slot)
{
  return ((unsigned char *) slot) - sizeof(struct fechdr);
}

static void eq_show_queue(eq_queue_t *que)
{
  printf("k:%d  n:%d  size:%d  totsize:%d  cnt:%d\n",
         que->k, que->n, que->size, que->totsize,
         que->cnt);
}

int eq_init_queue(eq_queue_t *que, int k, int n, int size)
{
  int i;

  DB("initializing FEC ...\n");

  que->fec = (void *) fec_new(k, n);  
  if (que->fec == NULL) return -1;

  DB("allocating memory ...\n");

  for (i=0; i<n; i++) {
    DB("malloc(que->buf[%d])\n", i);
    eq_alloc_slot((void **) &(que->buf[i]), size);
  }

  que->cnt = 0;
  que->k = k;
  que->n = n;
  que->size = size;
  que->totsize = size + sizeof(struct fechdr);

  eq_show_queue(que);

  DB("size:%d  rsphdr:%d  fechdr:%d\n",
     size, sizeof(struct rsphdr), sizeof(struct fechdr));

  return 0;
}

void eq_fini_queue(eq_queue_t *que)
{
  int i;

  DB("finializing FEC ...\n");

  if (que->fec != NULL) {
    fec_free(que->fec);
    que->fec = NULL;
  }

  DB("releasing memory ...\n");

  for (i = 0; i < que->n; i++) {
    DB("free(que->buf[%d])\n", i);
    eq_free_slot((void **) &(que->buf[i]));
  }

  que->cnt = 0;
  que->k = 0;
  que->n = 0;
  que->size = 0;
  que->totsize = 0;
}

int eq_enqueue_buffer(eq_queue_t *que, void *buf)
{
  if (que->cnt >= que->k)
    return -1;

  memcpy(que->buf[que->cnt], buf, que->size);
  que->cnt++;  
  return (que->cnt == que->k) ? 1 : 0;
}

inline void eq_flush_queue(eq_queue_t *que)
{
  que->cnt = 0;
}

void eq_encode_queue(eq_queue_t *que)
{
  int i;

  for (i=que->k; i<que->n; i++) {
    fec_encode(que->fec, (void *) que->buf, que->buf[i], i, que->size);
  }
}

void eq_add_fechdr(eq_queue_t *que, unsigned int seqno) {
  struct fechdr *fh;
  int i;

  for (i=0; i<que->n; i++) {
    fh = eq_get_slot_head(que->buf[i]);
    fh->k = que->k;
    fh->n = que->n;
    fh->index = i;
    fh->reserved = 0;
    fh->seqno = htonl(seqno);
    fh->size = htons(que->size);
    fh->checksum = htons(0);
  }
}
