#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <linux/ip.h>
#include <lib/util.h>
#include <router/ipqext/extensions/ipqext_fec/rawsock.h>

int create_raw_socket(struct raw_struct *raw)
{
  int on = 1;
  int sd = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);
  if (sd < 0) {
    perror("socket");
    return sd;
  }

  if (setsockopt(sd, IPPROTO_IP, IP_HDRINCL, (char *) &on, on) == -1) {
    perror("setsockopt");
    return -1;
  }

  raw->sd = sd;

  return sd;
}

void gen_raw_packet(struct raw_struct *raw)
{ 
  raw->iph.version = 4;
  raw->iph.ihl = RAW_IPH_LEN >> 2;
  raw->iph.tos = 0;
  raw->iph.tot_len = htons(raw->data_len);
  raw->iph.id = htons(getpid());
  raw->iph.frag_off = htons(0);
  raw->iph.ttl = MAXTTL;
  raw->iph.protocol = raw->protocol;
  raw->iph.check = 0;
  raw->iph.saddr = raw->src_addr.s_addr;
  raw->iph.daddr = raw->dst_addr.s_addr;
 
  raw->sockaddr.sin_family = AF_INET;
  raw->sockaddr.sin_addr.s_addr = raw->dst_addr.s_addr;
  raw->sockaddr.sin_port = htons(0);
}

void free_raw_packet(struct raw_struct *raw)
{
  if (raw->data != NULL) {
    free(raw->data);
    raw->data = NULL;
  }
}

int send_raw_packet(struct raw_struct *raw)
{
  struct msghdr msg;
  struct iovec iov[3];

  int retval;
  gen_raw_packet(raw);
  
  iov[0].iov_base = &(raw->iph);
  iov[0].iov_len = RAW_IPH_LEN;
  if (raw->protocol == IPPROTO_RSP) {
    iov[1].iov_base = &(raw->rsph);
    iov[1].iov_len = sizeof(struct rsphdr);
    iov[2].iov_base = raw->data;
    iov[2].iov_len = raw->data_len;
    msg.msg_iovlen = 3;
    DB(">> %x(%d)  %x(%d)  %x(%d)\n",
       iov[0].iov_base, iov[0].iov_len,
       iov[1].iov_base, iov[1].iov_len,
       iov[2].iov_base, iov[2].iov_len);
  } else {
    iov[1].iov_base = raw->data;
    iov[1].iov_len = raw->data_len;
    msg.msg_iovlen = 2;
    DB(">> %x(%d)  %x(%d)\n",
       iov[0].iov_base, iov[0].iov_len,
       iov[1].iov_base, iov[1].iov_len);
  }
  msg.msg_name = &(raw->sockaddr);
  msg.msg_namelen = sizeof(struct sockaddr);
  msg.msg_iov = (void *) &iov;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;

  retval = sendmsg(raw->sd, &msg, 0);

  if (retval < 0) {
    ERR();
  }
  return retval;
}

/*
void raw_socket_test() {
  struct raw_struct raw;

  memset(&raw, 0, sizeof(struct raw_struct));

  if (create_raw_socket(&raw) < 0)
    exit(-1);

  raw.data_len = 100;
  raw.data = malloc(raw.data_len);
  memset(raw.data, 'A', raw.data_len);
  raw.src_addr.s_addr = inet_addr("192.168.249.50");
  raw.dst_addr.s_addr = inet_addr("192.168.249.81");
  raw.protocol = IPPROTO_UDP;
  send_raw_packet(&raw);
  free_raw_packet(&raw);
  close(raw.sd);
}
*/
