// indicate this is an ipqext extension
#define _IPQEXT_EXTENSION_

//----------------------------------------------------------------------//
//
//  ipqext extension module traceback
//
//----------------------------------------------------------------------//

#include <linux/netfilter.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/ip.h>
#include <libipq.h>
#include <lib/util.h>
#include <sys/time.h>
#include <router/ipqext/ipqext.h>
#include <router/ddos/edgemark/edgemark.h>
#include <router/ddos/traceback/traceback.h>

/*
 * define extension name here
 *
 * extension name must match with the output module name
 * ( i.e. ipqext_{extension name} )
 */
const char extension_name[] = "traceback";

//----------------------------------------------------------------------//
//
//  module functions
//


enum state_t { STATE_INIT, STATE_LISTEN, STATE_FINISH };

#define MIN_CNT 100
#define MAX_CNT 5000
MarkingFields fields[MAX_CNT];

int fieldcnt = 0;
int param_cnt;
int param_calc;
double param_markprob;
unsigned long param_victim_addr;
enum state_t state = STATE_INIT;

int first, first_deta, first_gen_tree;
unsigned long elapsed;
struct timeval starttime, starttime_deta, endtime;

/*
 * ipqext structure
 */
static ipqext_t ext = {
  prev      : NULL,
  next      : NULL,
  handle    : NULL,
  active    : 0,
  name      : extension_name,
  version   : "1.0",
  groupmask : 0,
  init      : &init,
  fini      : &fini,
  proc      : &proc,
  start     : &start,
  stop      : &stop,
  reset     : &reset,
  info      : &info,
  help      : &help,
};


static void printLocalTrafficToFile(const RouterTree* head, double time)
{
  int i;
  static char cmd[500];
  struct in_addr in;
    
  in.s_addr = head->ip;
  snprintf(cmd, sizeof(cmd)-1, "echo \"%lf\t%lf\" >> _routerData_%s",
           time, head->localTrafficRate, inet_ntoa(in));
  system(cmd);
  for (i=0; i<head->numOfBranch; i++) {
    printLocalTrafficToFile(head->nextLevelArray[i], time);
  }
}

/*
 * packet processing function
 *
 * return: NF_ACCEPT   packet is accepted and leaves ipqext immediately
 *         NF_QUEUE    packet is queued for further ipqext processing
 *         NF_DROP     packet is explicitly dropped
 *
 *         negative verdict to indicate packet is altered
 */
int proc(ipq_packet_msg_t *m)
{
  unsigned int retval = NF_QUEUE;
  struct iphdr *iph;
  struct edgemark_opt *opt;
  int i;
  
  iph = (struct iphdr *) m->payload;

  if (state != STATE_LISTEN)
    goto end;
  
  if ((iph->ihl << 2) != EDGEMARK_IPHLEN)
    goto end;

  if (iph->daddr != param_victim_addr)
    goto end;
  
  opt = EDGEMARK_OPTPTR(iph);
  if ((opt->type != IPOPT_EDGEMARK) ||
      (opt->length != EDGEMARK_OPTLEN) ||
      (opt->msgtype != EDGEMARK_MSGTYPE_MARK))
    goto end;

  if (first == 0) {
    first = 1;
    gettimeofday(&starttime, NULL);
  }

  if (first_deta == 0) {
    first_deta = 1;
    gettimeofday(&starttime_deta, NULL);
  }
  
  fields[fieldcnt].startIP = opt->mark_start;
  // mark victim address as the end of the edge if end = 0x00000000
  fields[fieldcnt].endIP = (opt->mark_end == 0x00000000) ?
    param_victim_addr : opt->mark_end;
  fields[fieldcnt].distance = opt->mark_distance;
  

  /*DB*/ /*LOG(LOG_INFO, "mark[%d] st:%08lx ed:%08lx d:%ld\n",
             fieldcnt, fields[fieldcnt].startIP,
             fields[fieldcnt].endIP, fields[fieldcnt].distance);
         */
  
  fieldcnt++;
  if (fieldcnt == param_cnt) {
    // construct attack graph

    MarkingFieldsNode *fieldList = newMarkingFieldsNode();
    RouterTree* head = newRouterTree(param_victim_addr, 10);
    char* tree_str;
    unsigned long* sstree;
    int c, sstree_cnt;
    FILE *fp;
    struct in_addr in;
        
    // 1. add all the fields into a MarkingFieldsNode structure
    for (i=0; i<fieldcnt; i++) {
      addMarkingFields(fieldList, &fields[i]);
    }
    //printMarkingFieldsNodeList(fieldList);
    
    // 2. construct attack graph, stored in a RouterTree structure
    attackGraphCon(head, fieldList);
    //printTree(head, NULL);

    if (first_gen_tree == 0) {
      // 3. extract attack graph information for GUI
      tree_str = writeTreeStr(head);
      /*DB*/ LOG(LOG_INFO, "tree str = [%s]\n", tree_str);
      fp = fopen("_tree.dat", "w+");
      /*DB*/ LOG(LOG_INFO, "writing _tree.dat\n");
      if (fp != NULL) {
        //fwrite(tree_str, 1, strlen(tree_str)+1, fp); // this doesn't work, why?
        fprintf(fp, "%s\n", tree_str);
        fclose(fp);
      } else {
        CERR("Fail to write attack graph information to file.\n");
      }
      free(tree_str);
      first_gen_tree = 1;
    }
    
    // without local traffic rate calcuation
    if (param_calc == 0) {

      // 4. extract attack graph information for snapshot alogrithm
      sstree = writeTree(head);
      i = 0;
      /*DB*/ LOG(LOG_INFO, "sstree node list: ");
      for (sstree_cnt = 0; sstree[i] != 0xFFFFFFFF; i++) {
        in.s_addr = sstree[i];
        /*DB*/ LOG(LOG_INFO, "%s ", inet_ntoa(in));
        sstree_cnt++;
      }
      /*DB*/ LOG(LOG_INFO, "\n");
      /*DB*/ LOG(LOG_INFO, "sstree link lists:");
      for (c = 0; c <= sstree_cnt; i++) {
        if (sstree[i] == 0xFFFFFFFF) {
          /*DB*/ LOG(LOG_INFO, "\n");
          c++;
        } else {
          in.s_addr = sstree[i];
          /*DB*/ LOG(LOG_INFO, "%s ", inet_ntoa(in));
        }
      }

      /*DB*/ LOG(LOG_INFO, "writing _snapshot.dat\n");
      fp = fopen("_snapshot.dat", "w+");
      if (fp != NULL) {
        fwrite(sstree, sizeof(unsigned long), i, fp);
        fclose(fp);
      } else {
        CERR("Fail to write attack graph information to file.\n");
      }
    
      free(sstree);
      freeRouterTree(head);
      freeMarkingFieldsNode(fieldList);

      state = STATE_FINISH;
      goto end;
    } else {
      gettimeofday(&endtime, NULL);
      elapsed =
        (endtime.tv_sec - starttime_deta.tv_sec) * 1000000 +
        (endtime.tv_usec - starttime_deta.tv_usec);
      
      calLocalTrafficRate(head, fieldList, elapsed/1000, param_markprob);

      elapsed =
        (endtime.tv_sec - starttime.tv_sec) * 1000000 +
        (endtime.tv_usec - starttime.tv_usec);

      //printTree(head, NULL);
      printLocalTrafficToFile(head, elapsed / 1000000.0);
      fieldcnt = 0;
      first_deta = 0;
    }
  }
  
end:
  return retval;
}

/*
 * extension start function
 *
 * return on success, -1 or nonzero value on error
 */
int start()
{
  if (state != STATE_INIT)
    return -1;

  state = STATE_LISTEN;
  fieldcnt = 0;
  first = 0;
  first_deta = 0;
  first_gen_tree = 0;
  
  return 0;
}

/*
 * extension stop function
 *
 * return on success, -1 or nonzero value on error
 */
int stop()
{
  if (state == STATE_INIT)
    return -1;

  state = STATE_INIT;

  return 0;
}

/*
 * extension reset function
 *
 * return on success, -1 or nonzero value on error
 */
int reset()
{
  if (state == STATE_INIT)
    return -1;

  state = STATE_LISTEN;
  fieldcnt = 0;
  first = 0;
  first_deta = 0;
  first_gen_tree = 0;
  
  return 0;
}

/*
 * extension info function
 *
 * return a pointer to a char buffer or NULL
 */
const char infobuf[] = " info here ";

const char *info()
{
  return infobuf;
}

/*
 * extension help function
 *
 * return a pointer to a char buffer or NULL
 */
const char helpbuf[] = " help here ";

const char *help()
{
  DB("%s: help invoked\n", extension_name);
  return helpbuf;
}

/*
 *  initial pass function
 *
 * return on success, -1 or nonzero value on error
 */
int init_pass(void *handle, char *param, int paramlen)
{
  char temp[IPQEXT_PARAM_LEN];
  int i, st, ed, idx;
  struct in_addr in;
  
  // set module handle
  ext.handle = handle;
  
  // parse parameter buffer
  if ((param == NULL) || (paramlen == 0))
    return -1;

  st = 0;
  idx = 0;

  param_calc = 0;
  
  for (i=0; i<paramlen; i++) {
    if ((param[i] == ',') || (param[i] == '\0')) {
      ed = i;
      if (ed-st > 0) {
        strncpy(temp, &(param[st]), ed-st);
        temp[ed-st] = '\0';

        switch (idx) {
        case 0:
          param_victim_addr = inet_addr(temp);
          in.s_addr = param_victim_addr;
          LOG(LOG_INFO, "param: victim addr=%s\n", inet_ntoa(in));
          idx++;
          break;
          
        case 1:
          param_cnt = atoi(temp);
          
          if ((param_cnt < MIN_CNT) || (param_cnt > MAX_CNT)) {
            LOG(LOG_INFO, "param: cnt should be within [%d..%d]\n",
                MIN_CNT, MAX_CNT);
            return -1;
          }

          LOG(LOG_INFO, "param: cnt=%d\n", param_cnt);

          idx++;
          break;
          
        case 2:
          param_calc = (atoi(temp)==0) ? 0 : 1;
          LOG(LOG_INFO, "param: LTRCalc=%d\n", param_calc);
          idx++;
          break;

        case 3:
          param_markprob = atoi(temp) / 100.0;
          LOG(LOG_INFO, "param: markprob=%lf\n", param_markprob);

          return 0;
        }
      }
      st = i+1;
    }

    if (param[i] == '\0')
      break;
  }
  
  return (idx == 2) ? 0 : -1;
}

/*
 * extension initalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int init()
{
  // initalization codes here

  // register function must always be the last call
  ipqext_register(&ext);

  return 0;
}

/*
 * extension finalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int fini()
{
  // unregister function must always be the first call
  ipqext_unregister(&ext);

  // finialization codes here

  return 0;
}

/*
 * module initalization function
 */
void _init(void)
{
}

/*
 * module finalization function
 */
void _fini(void)
{
}

//----------------------------------------------------------------------//
