/*
 *  TODO:
 *  
 *  1) ext lock
 */

#define __DAEMON__

#ifdef __DAEMON__
#define __SYSLOG__
#endif

#include <asm/types.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <libipq.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <dlfcn.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <lib/util.h>
#include <ctype.h>
#include <stdarg.h>
#include <getopt.h>
#include <router/ipqext/ipqext.h>
#include <router/ipqext/ipqextctl.h>

#define DEFAULT_EXTPATH "extensions"   // default extension path
                                       // (relative to current directory)
#define EXTPATH_LEN     1024           // extension path length
#define BUFSIZE         2048           // packet buffer size
#define KRCVBUF_SIZE    (65536*20)     // kernel receive buffer size
#define KSNDBUF_SIZE    (65536*20)     // kernel send buffer size

// OBSOLETE
/*
static void die(struct ipq_handle *h)
{
  CERR("%s\n", ipq_errstr());
  close_all_extension_handles();
  ipqextctl_fini();
  ipq_destroy_handle(h);
  exit(1);
}
*/

/****************************************************************************
 *
 * extension chain handling routines
 *
 ****************************************************************************/

// IPQ extension header node
ipqext_t exthead = {
  prev   : &exthead,
  next   : &exthead,
  handle : NULL,
};

static char extension_path[EXTPATH_LEN+1];                 // extension path
static char lockfile_path[] = "/var/lock/subsys/ipqextd";  // lock file path
static char instance_path[40];                             // daemon instance
int locked = 0;

#define for_each_extension(head, ext) \
  for ((ext) = (head)->next; (ext) != (head); (ext) = (ext)->next)

static void *load_extension_handle(char *extname, char *param)
{
  char *error;
  char tmp[1024];
  void *handle;
  int (*init_pass) (void *handle, char *param, int paramlen);
  int (*init) ();

  snprintf(tmp, sizeof(tmp) - 1, "%s/ipqext_%s.so", extension_path,
           extname);
  LOG(LOG_INFO, "loading extension %s\n", tmp);

  handle = dlopen(tmp, RTLD_NOW);
  if ((error = dlerror()) != NULL) {
    CERR("%s\n", error);
    return NULL;
  }

  init_pass = dlsym(handle, "init_pass");

// OBSOLETE
/*
  if ((error = dlerror()) != NULL) {
    CERR("%s\n", error);

    dlclose(handle);
    if ((error = dlerror()) != NULL) {
      CERR("%s\n", error);
    }

    return NULL;
  } else {
*/

  if ((error = dlerror()) == NULL) {
    if (init_pass(handle, param, IPQEXT_PARAM_LEN) != 0) {
      CERR("init_pass failed\n");

      dlclose(handle);
      if ((error = dlerror()) != NULL) {
        CERR("%s\n", error);
      }

      return NULL;
    }
  }

  init = dlsym(handle, "init");
  if ((error = dlerror()) == NULL) {
    if (init() != 0) {
      CERR("initialization failed\n");

      dlclose(handle);
      if ((error = dlerror()) != NULL) {
        CERR("%s\n", error);
      }

      return NULL;
    }
  }

  return handle;
}

static inline void close_extension_handle(ipqext_t *ext)
{
  if (ext->fini != NULL) {
    if (ext->fini() != 0) {
      CERR("finalization failed\n");
    }
  }
  if (ext->handle != NULL) {
    dlclose(ext->handle);
  }
}

static void close_all_extension_handles()
{
  ipqext_t *ext, *next;

  LOG(LOG_INFO, "closing all extension handles ...\n");

  for (ext = exthead.next; ext != &exthead; ext = next) {
    next = ext->next;
    close_extension_handle(ext);
  }
}

static int get_extension_info(ipqext_t *ext, int idx, char *buf, int len)
{
  return snprintf(buf, len, "%5d   %s   %6lu    %s",
                  idx,
                  (ext->active == 0) ? "stopped  " : "started  ",
                  ext->groupmask, ext->name);
}

static int get_extension_mark(ipqext_t *head, char *list, char *buf,
                              int len)
{
  ipqext_t *ext;
  char *ptr;
  unsigned int mark = 0x0;
  char *listend = (list + strlen(list));
  char *start, *end;

  for_each_extension(head, ext) {
    ptr = strstr(list, ext->name);
    if (ptr != NULL) {
      start = (ptr - 1);
      end = (ptr + strlen(ext->name));

      if (((start == list - 1) || (*start == ',')) &&
          ((end == listend) || (*end == ','))) {
        mark |= ext->groupmask;
      }
    }
  }

  return snprintf(buf, len, "0x%x", mark);
}

static void add_extension(ipqext_t *head, ipqext_t *ext)
{
  ext->prev = head->prev;
  ext->next = head;
  ext->prev->next = ext;
  ext->next->prev = ext;
}

static void del_extension(ipqext_t *head, ipqext_t *ext)
{
  ext->prev->next = ext->next;
  ext->next->prev = ext->prev;
  ext->prev = NULL;
  ext->next = NULL;
}

static int move_extension(ipqext_t *head, ipqext_t *target, int idx)
{
  ipqext_t *ext;
  int i, ok, before;

  if (idx < 0)
    return -1;

  i = ok = before = 0;
  for (ext = head->next; ext != head; ext = ext->next, i++) {
    if (ext == target) {
      before = 1;
    }
    if (idx == i) {
      ok = 1;
      break;
    }
  }

  if (ok) {
    if (ext != target) {
      del_extension(head, target);
      if (before) {
        // if target is before ext, move target after ext
        target->prev = ext;
        target->next = ext->next;
        target->prev->next = target;
        target->next->prev = target;
      } else {
        // if target is after ext, move target before ext
        target->prev = ext->prev;
        target->next = ext;
        target->prev->next = target;
        target->next->prev = target;
      }
    }
    return 0;
  }

  return -1;
}

static ipqext_t *find_extension(char *extname)
{
  ipqext_t *ext;

  for_each_extension(&exthead, ext) {
    if (strncmp(ext->name, extname, IPQEXT_NAME_LEN) == 0) {
      return ext;
    }
  }
  return NULL;
}

void ipqext_register(ipqext_t *ext)
{
  LOG(LOG_INFO, "registering extension: %s (%s)\n", ext->name,
      ext->version);
  ext->active = 0;
  add_extension(&exthead, ext);
}

void ipqext_unregister(ipqext_t *ext)
{
  LOG(LOG_INFO, "unregistering extension: %s (%s)\n",
      ext->name, ext->version);
  del_extension(&exthead, ext);
}


/****************************************************************************
 *
 * ipqextctl (daemon-manager communication socket routines)
 *
 ****************************************************************************/

int ctlsd;
pthread_t ctl_thread;
pthread_mutex_t ipqext_lock = PTHREAD_MUTEX_INITIALIZER;
#define lock_ipqext(m) pthread_mutex_lock(m);
#define unlock_ipqext(m) pthread_mutex_unlock(m);

// send response to manager
static void ctl_response(int sd, int status, char *extname, char *param,
                         int paramlen)
{
  ipqextctl_response_t response;

  // prepare response
  memset(&response, 0, sizeof(ipqextctl_response_t));

  response.status = status;

  if (extname != NULL) {
    strncpy(response.extname, extname, IPQEXT_NAME_LEN);
  }

  if (param != NULL) {
    if (paramlen == -1)
      paramlen = strlen(param);
    memcpy(&(response.param), param, min(IPQEXT_PARAM_LEN, paramlen));
  }

  send(sd, &response, sizeof(ipqextctl_response_t), 0);
}

// request handling thread
void *ctl_thread_handle()
{

#define CHECK_EXT_EXISTED                                     \
  ext = find_extension(request.extname);                      \
  if (ext == NULL) {                                          \
    ctl_response(clisd, IPQEXTCTL_STAT_FAIL, request.extname, \
                 "Extension does not exist.", -1);            \
    break;                                                    \
  }

  struct sockaddr_un cliaddr;
  int clisd, cliaddrlen;
  int finished;
  int recvlen;
  ipqext_t *ext;
  ipqextctl_request_t request;
  int idx;
  char param[IPQEXT_PARAM_LEN + 1];
  int paramlen;
  char infoheader[] = "INDEX   STATUS   GROUPMASK    EXTENSION";
  const char *(*help) ();
  const char *infobuf, *helpbuf;

  pthread_detach(pthread_self());
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  while (1) {
    DB("waiting for connection ...\n");

    cliaddrlen = sizeof(struct sockaddr_un);
    clisd = accept(ctlsd, (struct sockaddr *) &cliaddr, &cliaddrlen);
    if (clisd == -1) {
      CERR("failed to accept\n");
      continue;
    }

    DB("accepted a connection ...\n");

    finished = 0;
    while (finished == 0) {
      recvlen = recv(clisd, &request, sizeof(ipqextctl_request_t), 0);
      if (recvlen == sizeof(ipqextctl_request_t)) {
        DB("received command: %d\n", request.command);

        switch (request.command) {
        case IPQEXTCTL_CMD_ADD:
          DB("executing command: add\n");

          ext = find_extension(request.extname);
          if (ext == NULL) {
            if (load_extension_handle
                (request.extname, request.param) != NULL) {
              ctl_response(clisd, IPQEXTCTL_STAT_OK,
                           request.extname, "Extension is added.", -1);
            } else {
              ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                           request.extname,
                           "Failed to load extension.", -1);
            }
          } else {
            ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                         request.extname, "Extension already exists.", -1);
          }
          break;

        case IPQEXTCTL_CMD_DEL:
          DB("executing command: delete\n");

          CHECK_EXT_EXISTED;

          close_extension_handle(ext);
          ctl_response(clisd, IPQEXTCTL_STAT_OK, request.extname,
                       "Extension is deleted.", -1);
          break;

        case IPQEXTCTL_CMD_MOVE:
          DB("executing command: move\n");

          idx = atoi(request.param);
          CHECK_EXT_EXISTED;

          if (move_extension(&exthead, ext, idx) == 0) {
            ctl_response(clisd, IPQEXTCTL_STAT_OK,
                         request.extname, NULL, 0);
          } else {
            ctl_response(clisd, IPQEXTCTL_STAT_OK,
                         request.extname, "Invalid index.", -1);
          }
          break;

        case IPQEXTCTL_CMD_GROUP:
          DB("executing command: group\n");

          idx = atoi(request.param);
          CHECK_EXT_EXISTED;

          ext->groupmask = idx;

          ctl_response(clisd, IPQEXTCTL_STAT_OK, request.extname, NULL, 0);
          break;

        case IPQEXTCTL_CMD_START:
          DB("executing command: start\n");

          CHECK_EXT_EXISTED;

          if (ext->active == 0) {
            if (ext->start != NULL) {
              if (ext->start() != 0) {
                ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                             request.extname,
                             "Failed to start extension.", -1);
                break;
              }
            }
            ext->active = 1;
            ctl_response(clisd, IPQEXTCTL_STAT_OK,
                         request.extname, "Extension is started.", -1);
          } else {
            ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                         request.extname,
                         "Extension has already been started.", -1);
          }
          break;

        case IPQEXTCTL_CMD_STOP:
          DB("executing command: stop\n");

          CHECK_EXT_EXISTED;

          if (ext->active != 0) {
            if (ext->stop != NULL) {
              if (ext->stop() != 0) {
                ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                             request.extname,
                             "Failed to stop extension.", -1);
                break;
              }
            }
            ext->active = 0;
            ctl_response(clisd, IPQEXTCTL_STAT_OK,
                         request.extname, "Extension is stopped.", -1);
          } else {
            ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                         request.extname,
                         "Extension has not been started.", -1);
          }
          break;

        case IPQEXTCTL_CMD_RESET:
          DB("executing command: reset\n");

          if (request.extname[0] == '\0') {
            ctl_response(clisd, IPQEXTCTL_STAT_OK, NULL,
                         "All extensions are reset.", -1);
          } else {
            CHECK_EXT_EXISTED;

            if (ext->reset != NULL) {
              ext->reset();
            }
            ctl_response(clisd, IPQEXTCTL_STAT_OK,
                         request.extname, "Extension is reset.", -1);
          }
          break;

        case IPQEXTCTL_CMD_INFO:
          DB("executing command: info\n");

          CHECK_EXT_EXISTED;

          if ((ext->info != NULL)
              && ((infobuf = ext->info()) != NULL)) {
            ctl_response(clisd, IPQEXTCTL_STAT_OK,
                         request.extname, (char *) infobuf, -1);
          } else {
            ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                         request.extname, "No information available.", -1);
          }
          break;

        case IPQEXTCTL_CMD_FLUSH:
          DB("executing command: flush\n");

          close_all_extension_handles();
          ctl_response(clisd, IPQEXTCTL_STAT_OK, request.extname,
                       "All extensions are deleted.", -1);
          break;

        case IPQEXTCTL_CMD_LIST:
          DB("executing command: list\n");

          idx = 0;
          paramlen = min(IPQEXT_PARAM_LEN, strlen(infoheader));
          ctl_response(clisd, IPQEXTCTL_STAT_CONT, NULL,
                       infoheader, paramlen);

          for_each_extension(&exthead, ext) {
            paramlen =
                get_extension_info(ext, idx++, param, IPQEXT_PARAM_LEN);
            ctl_response(clisd, IPQEXTCTL_STAT_CONT, request.extname,
                         param, paramlen);
          }
          if (idx == 0)
            paramlen = snprintf(param, IPQEXT_PARAM_LEN, "No extension.");
          else
            paramlen = snprintf(param, IPQEXT_PARAM_LEN,
                                "Total %d extension(s).", idx);
          ctl_response(clisd, IPQEXTCTL_STAT_OK, NULL, param, paramlen);
          break;

        case IPQEXTCTL_CMD_MARK:
          DB("executing command: mark\n");

          // default: no error response, reply 0x0 on error
          paramlen = get_extension_mark(&exthead, request.param,
                                        param, IPQEXT_PARAM_LEN);
          ctl_response(clisd, IPQEXTCTL_STAT_OK, NULL, param, paramlen);
          break;

        case IPQEXTCTL_CMD_HELP:
          {
            char *error;
            char tmp[1024];
            void *handle;

            DB("executing command: help\n");

            handle = NULL;
            help = NULL;

            ext = find_extension(request.extname);
            if (ext == NULL) {

              snprintf(tmp, sizeof(tmp) - 1,
                       "%s/ipqext_%s.so", extension_path, request.extname);
              LOG(LOG_INFO, "loading extension %s\n", tmp);

              handle = dlopen(tmp, RTLD_NOW);
              if ((error = dlerror()) != NULL) {
                CERR("%s\n", error);
                handle = NULL;
              } else {
                help = dlsym(handle, "help");
                if ((error = dlerror()) != NULL) {
                  CERR("%s\n", error);
                  help = NULL;
                }
              }

            } else {
              help = ext->help;
            }

            if ((help != NULL) && ((helpbuf = help()) != NULL)) {
              ctl_response(clisd, IPQEXTCTL_STAT_OK,
                           request.extname, (char *) helpbuf, -1);
            } else {
              ctl_response(clisd, IPQEXTCTL_STAT_FAIL,
                           request.extname,
                           "No help information available.", -1);
            }

            if (handle != NULL) {
              dlclose(handle);
              if ((error = dlerror()) != NULL) {
                CERR("%s\n", error);
              }
            }
          }
          break;

        default:
          DB("unknown command\n");

          ctl_response(clisd, IPQEXTCTL_STAT_UNKNOWN,
                       request.extname, "Unknown command.", -1);
        }

      } else {
        DB("received %d byte(s)\n", recvlen);
        close(clisd);
        finished = 1;
        DB("connection closed ...\n");
      }
    }
  }

  pthread_exit(NULL);
  return NULL;

#undef CHECK_EXT_EXISTED
}

static inline int ipqextctl_init()
{
  LOG(LOG_INFO, "initializing ipqextctl ...\n");

  ctlsd = IPQEXTCTL_INIT_TCP_SOCKET(IPQEXTD_SOCKET_PATHNAME);
  if (ctlsd == -1) {
    return -1;
  }

  if (listen(ctlsd, 3) != 0) {
    ERR_RET(-1);
  }

  pthread_create(&ctl_thread, NULL, ctl_thread_handle, NULL);

  return 0;
}

static inline void ipqextctl_fini()
{
  LOG(LOG_INFO, "finalizing ipqextctl ...\n");

  pthread_cancel(ctl_thread);
  close(ctlsd);
  ipqextctl_close_socket(ctlsd, IPQEXTD_SOCKET_PATHNAME);
}


/****************************************************************************
 *
 * ipqueue
 *
 ****************************************************************************/

struct ipq_handle *ipqhandle;

static inline int ipqueue_init()
{
  int status;

  int ret;
  unsigned long bufsize, l;

  LOG(LOG_INFO, "initializing ipqueue handle ...\n");

  if (!(ipqhandle = ipq_create_handle(0, PF_INET))) {
    ipq_destroy_handle(ipqhandle);
    CERR_RET(-1, "%s: %s\n", ipq_errstr(), strerror(errno));
  }
  //  if ((status = ipq_set_mode(ipqhandle, IPQ_COPY_META, BUFSIZE)) < 0) {  
  if ((status = ipq_set_mode(ipqhandle, IPQ_COPY_PACKET, BUFSIZE)) < 0) {
    ipq_destroy_handle(ipqhandle);
    CERR_RET(-1, "%s: %s\n", ipq_errstr(), strerror(errno));
  }

  /* set kernel receive buffer size */
  LOG(LOG_INFO, "initializing kernel RCVBUF size ...\n");

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(ipqhandle->fd, SOL_SOCKET, SO_RCVBUF,
                   &bufsize, (socklen_t *) & l);
  LOG(LOG_INFO, "current RCVBUF size is %lu\n", bufsize);

  bufsize = KRCVBUF_SIZE;
  l = (socklen_t) sizeof(bufsize);
  ret = setsockopt(ipqhandle->fd, SOL_SOCKET, SO_RCVBUF, &bufsize, l);
  LOG(LOG_INFO, "changing RCVBUF size to %lu\n", bufsize);

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(ipqhandle->fd, SOL_SOCKET, SO_RCVBUF,
                   &bufsize, (socklen_t *) & l);
  LOG(LOG_INFO, "new RCVBUF size is %lu\n", bufsize);


  /* set kernel send buffer size */
  LOG(LOG_INFO, "initializing kernel RCVBUF size ...\n");

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(ipqhandle->fd, SOL_SOCKET, SO_SNDBUF,
                   &bufsize, (socklen_t *) & l);
  LOG(LOG_INFO, "current SNDBUF size is %lu\n", bufsize);

  bufsize = KSNDBUF_SIZE;
  l = (socklen_t) sizeof(bufsize);
  ret = setsockopt(ipqhandle->fd, SOL_SOCKET, SO_SNDBUF, &bufsize, l);
  LOG(LOG_INFO, "changing SNDBUF size to %lu\n", bufsize);

  l = (socklen_t) sizeof(bufsize);
  ret = getsockopt(ipqhandle->fd, SOL_SOCKET, SO_SNDBUF,
                   &bufsize, (socklen_t *) & l);
  LOG(LOG_INFO, "new SNDBUF size is %lu\n", bufsize);

  LOG(LOG_INFO, "extension path: %s\n\n", extension_path);

  return 0;
}


/****************************************************************************
 *
 * ipqextd
 *
 ****************************************************************************/

static inline void ipqextd_init()
{
  openlog("ipqextd", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_LOCAL1);

  // initialize ipqueue
  if (ipqueue_init() == -1) {
    closelog();
    exit(1);
  }
  // initialize ipqext control interface
  if (ipqextctl_init() == -1) {
    ipq_destroy_handle(ipqhandle);
    closelog();
    exit(1);
  }
}

static inline void ipqextd_fini()
{
  close_all_extension_handles();
  ipqextctl_fini();
  ipq_destroy_handle(ipqhandle);

  closelog();
}

// daemon main handling routine
static void ipqextd_main()
{
  int status;
  unsigned char buf[BUFSIZE];
  struct ipqext *ext;
  int verdict;
  int altered;

  do {
    if ((status = ipq_read(ipqhandle, buf, BUFSIZE, 0)) < 0) {
      CERR("%s: %s\n", ipq_errstr(), strerror(errno));
      continue;
      // die(ipqhandle); // OBSOLETE
    }

    switch (ipq_message_type(buf)) {
    case NLMSG_ERROR:
      CERR("Received error message %d\n", ipq_get_msgerr(buf));
      break;

    case IPQM_PACKET:
      {
        ipq_packet_msg_t *m = ipq_get_packet(buf);

        DB("packet mark = %lu\n", m->mark);

        verdict = NF_ACCEPT;
        altered = 0;

        // passing through extension chain
        for_each_extension(&exthead, ext) {
          if ((ext->active != 0) && ((m->mark & ext->groupmask) > 0)) {
            verdict = ext->proc(m);

            // check if packet is altered
            if (verdict < 0) {
              altered = 1;
              verdict = -verdict;
            }
            // check verdict value
            if (verdict != NF_QUEUE)
              break;
          }
        }

        if (verdict == NF_QUEUE)
          verdict = NF_ACCEPT;

        // set verdict and replace packet content when necessary
        if (altered == 0) {
          status =
              ipq_set_verdict(ipqhandle, m->packet_id, verdict, 0, NULL);
        } else {
          status =
              ipq_set_verdict(ipqhandle, m->packet_id, verdict,
                              m->data_len, m->payload);
        }

        if (status < 0) {
          // die(ipqhandle); // OBSOLETE
          CERR("%s: %s\n", ipq_errstr(), strerror(errno));
          continue;
        }
        break;
      }

    default:
      CERR("Unknown message type!\n");
      break;
    }
  }
  while (1);
}

static void signal_handler(int signo)
{
  switch (signo) {
  case SIGQUIT:
  case SIGINT:
  case SIGHUP:
    ipqextd_fini();
    exit(0);
    break;
  }
}

static void catch_signal()
{
  signal(SIGQUIT, signal_handler);
  signal(SIGINT, signal_handler);
  signal(SIGHUP, signal_handler);
}

static void exit_handler()
{
  if (locked == 1) {
    if (unlink(instance_path))
      ERR();
  }
}

// daemon instance locking
static int init_lock()
{
  struct stat stat_buf;
  int pid = getpid();

  sprintf(instance_path, "/tmp/ipqextd.%d", pid);
  DB("instance pid = %d\n", pid);
  if (link(lockfile_path, instance_path) == -1) {
    CERR_RET(-1, "Failed to acquire lock at %s\n", lockfile_path);
  }
  if (stat(lockfile_path, &stat_buf) == -1) {
    ERR();
    goto exit_unlink;
  }
  if (stat_buf.st_nlink > 2) {
    LOG(LOG_ERR, "Another ipqextd daemon is already running!\n");
    goto exit_unlink;
  }

  goto exit_normal;

exit_unlink:
  if (unlink(instance_path))
    ERR();
  return -1;

exit_normal:
  locked = 1;
  return 0;
}


/****************************************************************************
 *
 * main
 *
 ****************************************************************************/

// program information
const char *program_version = "0.1";
const char *program_name = "ipqextd";


// options
static struct option opts[] = {
  {"extpath", required_argument, 0, 'e'},
  {"help", no_argument, 0, 'h'},
  {0, 0, 0, 0}
};

static const char cmdflags[] = { 'e', 'h' };

// error codes
#define ERR_INVALID_CMD 1

// exit messages
static void exit_tryhelp()
{
  fprintf(stderr, "Try `%s -h' or '%s --help' for more information.\n",
          program_name, program_name);
}

static void exit_error(int status, char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fprintf(stderr, "%s v%s: ", program_name, program_version);
  if (msg != NULL)
    vfprintf(stderr, msg, args);
  va_end(args);
  fprintf(stderr, "\n");

  if (status == ERR_INVALID_CMD) {
    exit_tryhelp();
  }

  exit(status);
}

static void exit_printhelp(void)
{
  printf("%s v%s\n\n"
         "Usage: %s [-e path]\n"
         "       %s [-h]\n"
         "\n", program_name, program_version, program_name, program_name);

  printf("Commands:\n"
         "Either long or short options are allowed.\n"
         "  --extpath -e path               Specify the extension path\n"
         "  --help    -h                    Print this help\n" "\n");
  exit(0);
}

// parse options in the command arguments
static void parse_options(int argc, char **argv)
{
  int optidx = 0;
  int c;
  int len;
  
  memset(extension_path, 0, sizeof(extension_path));
  if (getcwd(extension_path, EXTPATH_LEN-strlen(DEFAULT_EXTPATH)-1) == NULL) {
    strcpy(extension_path, "./");
    len = 2;
  } else {
    len = strlen(extension_path)+1;
    extension_path[len-1] = '/';
  }
  strncat(extension_path, DEFAULT_EXTPATH, EXTPATH_LEN - len);

  opterr = 0;
  while ((c = getopt_long(argc, argv, "e:h", opts, &optidx)) != -1) {
    switch (c) {
    case 0:
      break;

    case 'e':
      if (optind <= argc) {
        int len = min(EXTPATH_LEN - 1, strlen(argv[optind - 1]));
        strncpy(extension_path, argv[optind - 1], len);
        extension_path[len] = '\0';
      } else {
        exit_error(ERR_INVALID_CMD, "Invalid argument.");
      }
      break;

    case 'h':
      exit_printhelp();
      break;

    case '?':
      if (isprint(optopt))
        exit_error(ERR_INVALID_CMD,
                   "Unknown option or missing argument for switch `-%c'.",
                   optopt);
      else
        exit_error(ERR_INVALID_CMD,
                   "Unknown option character `\\x%x'.", optopt);
      break;

    default:
      exit_error(ERR_INVALID_CMD, "Invalid command.");
    }
  }
}

int main(int argc, char **argv)
{
  parse_options(argc, argv);

#ifdef __DAEMON__
  if (daemon(0, 0) == -1) {
    fprintf(stderr, "%s", strerror(errno));
    exit(1);
  }
#endif

  atexit(&exit_handler);
  if (init_lock() == -1) {
    exit(1);
  }
  catch_signal();

  ipqextd_init();
  ipqextd_main();
  ipqextd_fini();
  return 0;
}
